import QtQuick 2.4
import Ubuntu.Components 1.2
import QtHalide 0.1

Item {
    id: effectsGrid

    property ListModel availableEffects
    property HalideImage originalImage
    property HalideTransform halideTransform
    property var currentModel
    property var currentLocalFunction
    property var currentGlobalFunction

    property real cellSize: units.gu(15)
    property real maximumHeightRatio: 0.5
    property real maximumHeight: parent.height * maximumHeightRatio
    property real allCellsWidth: gridView.count * cellSize
    property int maximumRows: Math.max(Math.min(Math.floor(allCellsWidth / effectsGrid.width), 2), 1)
    property int rows: Math.min(Math.floor(maximumHeight / cellSize), maximumRows)
    height: rows * cellSize

    function createFunction(functionQml) {
        var functionComponent = Qt.createComponent(functionQml);
        return functionComponent.createObject();
    }

    function addFunction(functionObject) {
        halideTransform.functions.append(functionObject);
    }

    function removeFunction(functionObject) {
        halideTransform.functions.remove(functionObject);
    }

    GridView {
        id: gridView

        anchors.fill: parent
        model: availableEffects
        flow: GridView.FlowTopToBottom
        cellWidth: cellHeight
        cellHeight: gridView.height / rows
        delegate: Item {
            id: effect

            property bool activated: false
            property bool editing: effectsGrid.currentModel === model
            property HalideFunction globalFunction: createFunction(model.functionQml)
            property HalideFunction localFunction: createFunction(model.functionQml)

            width: gridView.cellWidth
            height: gridView.cellHeight

            HalideTransform {
                id: halideTransform
                live: true
                input: originalImage
                functions: [localFunction]
            }

            HalideImageRenderer {
                anchors.fill: parent
                image: halideTransform.output
                fillMode: HalideImageRenderer.PreserveAspectCrop
                verticalAlignment: HalideImageRenderer.AlignTop
                clip: true
                opacity: activated ? 1.0 : 0.5
                Behavior on opacity { UbuntuNumberAnimation {duration: UbuntuAnimation.SnapDuration} }
            }

            function activate() {
                if (!effect.activated) {
                    addFunction(effect.globalFunction);
                    effect.activated = true;
                }
            }

            function deactivate() {
                if (effect.activated) {
                    removeFunction(effect.globalFunction);
                    effect.activated = false;
                    if (editing) {
                        exitEdit();
                    }
                }
            }

            function enterEdit() {
                effectsGrid.currentLocalFunction = effect.localFunction;
                effectsGrid.currentGlobalFunction = effect.globalFunction;
                effectsGrid.currentModel = model;
            }

            function exitEdit() {
                effectsGrid.currentLocalFunction = null;
                effectsGrid.currentGlobalFunction = null;
                effectsGrid.currentModel = null;
            }

            function toggle() {
                if (effect.activated) {
                    deactivate();
                } else {
                    activate();
                }
            }

            function toggleIfEditing() {
                if (effect.activated && effect.editing) {
                    deactivate();
                } else {
                    activate();
                    enterEdit();
                }
            }

            MouseArea {
                anchors.fill: parent
                acceptedButtons: Qt.LeftButton | Qt.RightButton
                onClicked: mouse.button == Qt.LeftButton ? toggleIfEditing() : toggle()
            }

            Item {
                anchors {
                    right: parent.right
                    left: parent.left
                    bottom: parent.bottom
                }
                height: units.gu(4)

                Rectangle {
                    anchors.fill: parent
                    color: effect.activated ? UbuntuColors.green : "black"
                    opacity: 0.75
                    Behavior on opacity { ColorAnimation { duration: UbuntuAnimation.SnapDuration} }
                }

                Label {
                    id: label
                    anchors.centerIn: parent
                    text: model.label
                    style: Text.Raised
                }
            }
        }
    }
}
