import QtQuick 2.4
import Ubuntu.Components 1.2
import Ubuntu.Components.Popups 1.0
import QtHalide 0.1

Rectangle {
    id: overlay

    property ListModel availableEffects
    property HalideTransform liveTransform
    property HalideTransform saveTransform
    property url source

    width: units.gu(30)
    color: Qt.rgba(0.0, 0.0, 0.0, 0.9)

    function addFunction(functionQml) {
        var functionComponent = Qt.createComponent(functionQml);
        var functionObject = functionComponent.createObject();
        liveTransform.functions.append(functionObject);
        appliedEffectsView.positionViewAtEnd();
    }

    Column {
        id: column
        spacing: units.gu(2)
        anchors {
            fill: parent
            margins: units.gu(2)
        }

        ShapeButton {
            id: saveButton
            anchors {
                left: parent.left
                right: parent.right
            }

            property string outputFile: "output.png"
            text: "Save to %1".arg(outputFile)
            onClicked: editor.saveToFile(outputFile)

            ActivityIndicator {
                anchors {
                    verticalCenter: parent.verticalCenter
                    right: parent.right
                    rightMargin: units.gu(1)
                }
                width: units.gu(2)
                height: width
                running: saveTransform.updating
                opacity: running ? 1.0 : 0.0
                Behavior on opacity { UbuntuNumberAnimation {duration: UbuntuAnimation.BriskDuration} }
            }
        }

        Item {
            id: addEffect
            anchors {
                left: parent.left
                right: parent.right
            }
            height: addButton.height
            property bool opened: false

            ShapeButton {
                id: addButton
                anchors {
                    left: parent.left
                    right: parent.right
                }
                visible: !addEffect.opened
                text: "Add Effect"
                onClicked: addEffect.opened = true
            }

            UbuntuShape {
                id: effectsSelector
                anchors {
                    left: parent.left
                    right: parent.right
                    margins: units.gu(2)
                }
                parent: overlay
                y: addEffect.y + column.y
                height: addEffect.opened ? effectsList.count * units.gu(5) : addButton.height
                Behavior on height { UbuntuNumberAnimation {duration: UbuntuAnimation.FastDuration} }
                clip: true
                radius: "medium"
                backgroundColor: addEffect.opened ? Qt.rgba(0.06, 0.06, 0.06, 0.97) : "transparent"

                ListView {
                    id: effectsList
                    anchors.fill: parent
                    model: availableEffects
                    interactive: false

                    delegate: ListItem {
                        id: effectButton
                        height: units.gu(5)
                        visible: addEffect.opened
                        Label {
                            text: model.label
                            color: "white"
                            anchors.centerIn: parent
                        }
                        highlightColor: Qt.rgba(1, 1, 1, 0.10)
                        enabled: addEffect.opened
                        onClicked: {
                            addEffect.opened = false;
                            overlay.addFunction(model.functionQml);
                        }
                    }
                }
            }
        }

        ShapeButton {
            id: clearButton
            anchors {
                left: parent.left
                right: parent.right
            }

            text: "Remove All"
            onClicked: liveTransform.destroyFunctions()
            opacity: liveTransform.functions.count > 0 ? 1.0 : 0.0
            Behavior on opacity { UbuntuNumberAnimation {duration: UbuntuAnimation.BriskDuration} }
            visible: opacity != 0.0
        }

        ShapeButton {
            id: reorderButton
            anchors {
                left: parent.left
                right: parent.right
            }

            text: appliedEffectsView.ViewItems.dragMode ? "Done" : "Reorder"
            onClicked: appliedEffectsView.ViewItems.dragMode = !appliedEffectsView.ViewItems.dragMode
            backgroundColor: appliedEffectsView.ViewItems.dragMode ? (reorderButton.pressed ? Qt.rgba(0, 1, 0, 0.65) : Qt.rgba(0, 1, 0, 0.45)) :
                                                                     (reorderButton.pressed ? Qt.rgba(1, 1, 1, 0.15) : Qt.rgba(1, 1, 1, 0.05))
            opacity: liveTransform.functions.count > 0 ? 1.0 : 0.0
            Behavior on opacity { UbuntuNumberAnimation {duration: UbuntuAnimation.BriskDuration} }
            visible: opacity != 0.0
        }

        ListView {
            id: appliedEffectsView
            anchors {
                left: parent.left
                right: parent.right
            }

            UbuntuShape {
                anchors.fill: parent
                radius: "medium"
                backgroundColor: Qt.rgba(1, 1, 1, 0.05)
            }

            ActivityIndicator {
                anchors {
                    top: parent.top
                    topMargin: units.gu(2)
                    right: parent.right
                    rightMargin: units.gu(1)
                }
                width: units.gu(2)
                height: width
                running: liveTransform.updating
                opacity: running ? 1.0 : 0.0
                Behavior on opacity { UbuntuNumberAnimation {duration: UbuntuAnimation.BriskDuration} }
            }

            clip: true
            height: Math.min(units.gu(50), contentItem.height)
            Behavior on height { UbuntuNumberAnimation {duration: UbuntuAnimation.BriskDuration} }
            model: overlay.liveTransform.functions
            verticalLayoutDirection: ListView.BottomToTop
            displaced: Transition {
                UbuntuNumberAnimation { properties: "y" }
            }
            ViewItems.onDragUpdated: {
                // Live update as you drag
                if (event.status == ListItemDrag.Moving) {
                    liveTransform.functions.move(event.from, event.to, 1);
                }
            }

            delegate: ListItem {
                id: effectDelegate

                height: units.gu(5) + transformPreviewLoader.height
                divider {
                    visible: false
                }
                onPressAndHold: appliedEffectsView.ViewItems.dragMode = !appliedEffectsView.ViewItems.dragMode
                onClicked: modelData.enabled = !modelData.enabled
                contentItem.clip: false

                leadingActions: actions
                trailingActions: actions
                ListItemActions {
                    id: actions
                    actions: [
                        Action {
                            iconName: "delete"
                            onTriggered: {
                                var func = modelData;
                                liveTransform.functions.remove(func);
                                func.destroy();
                            }
                        }
                    ]
                }

                Item {
                    anchors {
                        fill: parent
                        topMargin: units.gu(2)
                        leftMargin: units.gu(2)
                        rightMargin: units.gu(2)
                    }

                    Label {
                        id: label
                        anchors {
                            top: parent.top
                            left: parent.left
                            right: parent.right
                        }
                        text: availableEffects.findElementWithProperty("name", modelData.name).label
                        font.bold: true
                        opacity: modelData.enabled ? 1.0 : 0.3
                    }

                    Image {
                        id: disabledIcon
                        anchors {
                            top: parent.top
                            right: parent.right
                        }
                        width: units.gu(2.5)
                        height: width
                        fillMode: Image.PreserveAspectFit
                        source: "disabled.png"
                        visible: !modelData.enabled
                    }

                    Loader {
                        id: transformPreviewLoader
                        anchors {
                            left: parent.left
                            right: parent.right
                            bottom: parent.bottom
                        }
                        opacity: modelData.enabled ? 1.0 : 0.3

                        source: availableEffects.findElementWithProperty("name", modelData.name).editor
                        onLoaded: transformPreviewLoader.item.setFunction(modelData)
                    }
                }
            }
        }
    }
}
