#include <Halide.h>

int custom_generate_filter_main(int argc, char **argv, std::ostream &cerr);

int main(int argc, char **argv) {
    return custom_generate_filter_main(argc, argv, std::cerr);
}

using namespace Halide::Internal;

// Copied from Halide (src/Generator.cpp) and tweaked to add the -b parameter
int custom_generate_filter_main(int argc, char **argv, std::ostream &cerr) {
    const char kUsage[] = "gengen [-g GENERATOR_NAME] [-f FUNCTION_NAME] [-b OUTPUT_FILE_NAME] [-o OUTPUT_DIR] [-e EMIT_OPTIONS] "
                          "target=target-string [generator_arg=value [...]]\n\n"
                          "  -e  A comma separated list of optional files to emit. Accepted values are "
                          "[assembly, bitcode, stmt, html]\n";

    std::map<std::string, std::string> flags_info = { { "-f", "" }, { "-b", "" }, { "-g", "" }, { "-o", "" }, { "-e", "" } };
    std::map<std::string, std::string> generator_args;

    for (int i = 1; i < argc; ++i) {
        if (argv[i][0] != '-') {
            std::vector<std::string> v = split_string(argv[i], "=");
            if (v.size() != 2 || v[0].empty() || v[1].empty()) {
                cerr << kUsage;
                return 1;
            }
            generator_args[v[0]] = v[1];
            continue;
        }
        auto it = flags_info.find(argv[i]);
        if (it != flags_info.end()) {
            if (i + 1 >= argc) {
                cerr << kUsage;
                return 1;
            }
            it->second = argv[i + 1];
            ++i;
            continue;
        }
        cerr << "Unknown flag: " << argv[i] << "\n";
        cerr << kUsage;
        return 1;
    }

    std::vector<std::string> generator_names = GeneratorRegistry::enumerate();
    if (generator_names.size() == 0) {
        cerr << "No generators have been registered\n";
        cerr << kUsage;
        return 1;
    }

    std::string generator_name = flags_info["-g"];
    if (generator_name.empty()) {
        // If -g isn't specified, but there's only one generator registered, just use that one.
        if (generator_names.size() != 1) {
            cerr << "-g must be specified if multiple generators are registered:\n";
            for (auto name : generator_names) {
                cerr << "    " << name << "\n";
            }
            cerr << kUsage;
            return 1;
        }
        generator_name = generator_names[0];
    }
    std::string function_name = flags_info["-f"];
    if (function_name.empty()) {
        // If -f isn't specified, assume function name = generator name.
        function_name = generator_name;
    }
    std::string output_file_name = flags_info["-b"];
    if (output_file_name.empty()) {
        // If -b isn't specified, assume file name = function name.
        output_file_name = function_name;
    }
    std::string output_dir = flags_info["-o"];
    if (output_dir.empty()) {
        cerr << "-o must always be specified.\n";
        cerr << kUsage;
        return 1;
    }
    if (generator_args.find("target") == generator_args.end()) {
        cerr << "Target missing\n";
        cerr << kUsage;
        return 1;
    }
    GeneratorBase::EmitOptions emit_options;
    std::vector<std::string> emit_flags = split_string(flags_info["-e"], ",");
    for (const std::string &opt : emit_flags) {
        if (opt == "assembly") {
            emit_options.emit_assembly = true;
        } else if (opt == "bitcode") {
            emit_options.emit_bitcode = true;
        } else if (opt == "stmt") {
            emit_options.emit_stmt = true;
        } else if (opt == "html") {
            emit_options.emit_stmt_html = true;
        } else if (!opt.empty()) {
            cerr << "Unrecognized emit option: " << opt
                 << " not one of [assembly, bitcode, stmt, html], ignoring.\n";
        }
    }

    std::unique_ptr<GeneratorBase> gen = GeneratorRegistry::create(generator_name, generator_args);
    if (gen == nullptr) {
        cerr << "Unknown generator: " << generator_name << "\n";
        cerr << kUsage;
        return 1;
    }
    gen->emit_filter(output_dir, function_name, output_file_name, emit_options);
    return 0;
}
