#include <Halide.h>
#include "halide_generators_common.h"

namespace {

const float MAX_EFFECT_SHIFT = 0.5f;
const float MIN_TONAL_WIDTH = 0.1f;
const float MAX_TONAL_WIDTH = 1.0f;
const float TONAL_WIDTH = 1.0f;
const int SHADOW_DETECT_MIN_INTENSITY = 2;
const int SHADOW_DETECT_MAX_INTENSITY = 90;
const int SHADOW_DETECT_INTENSITY_RANGE = SHADOW_DETECT_MAX_INTENSITY - SHADOW_DETECT_MIN_INTENSITY;
const float SHADOW_AGGRESSIVENESS_MUL = 0.45f;

const float DEFAULT_LOW_DISCARD_MASS = 0.02f;
const float DEFAULT_HIGH_DISCARD_MASS = 0.98f;

class AutoEnhance : public HalideGeneratorExtended<AutoEnhance> {
public:
    ImageParam input{ Halide::type_of<uint8_t>(), 3, "input" };

    Expr hermiteGamma(Expr x, Expr user_interval_upper) {
        Expr nonzeroIntervalUpper = clamp(user_interval_upper, 0.1f, 1.0f);
        Expr xScale = 1.0f / nonzeroIntervalUpper;
        Expr indep_var = xScale * x;
        Expr dep_var =  6.0f * ((indep_var * indep_var * indep_var) - (2.0f * (indep_var * indep_var)) + (indep_var));
        return Halide::select(x < 0.0f || x > nonzeroIntervalUpper, 0.0f,
                              clamp(dep_var, 0.0f, 1.0f));
    }

    Expr shadowDetailRemap(Expr i, Expr intensity) {
        Expr fi = i / 255.0f;
        Expr effect_shift = MAX_EFFECT_SHIFT * intensity;
        Expr weight = hermiteGamma(fi, TONAL_WIDTH);
        Expr remapped = 255.0f * (weight * (fi + effect_shift)) + (1.0f - weight) * fi + 0.5f;
        return remapped;
    }

    Expr toneExpansionRemap(Expr i, Expr low_discard_mass, Expr high_discard_mass, Func cumulProb) {
        Var j;
        RDom shr(0, 255);

        Func lowKink("lowKink");
        lowKink(j) = cast<uint8_t>(0);
        lowKink(0) = Halide::maximum(Halide::select(cumulProb(shr) < low_discard_mass, cast<uint8_t>(shr), cast<uint8_t>(0)));

        Func highKink("highKink");
        highKink(j) = cast<uint8_t>(0);
        highKink(0) = Halide::minimum(Halide::select(cumulProb(shr) > high_discard_mass, cast<uint8_t>(shr), cast<uint8_t>(255)));

        // REMAP
        Expr low_kink_f = cast<float>(lowKink(0)) / 255.0f;
        Expr high_kink_f = cast<float>(highKink(0)) / 255.0f;

        Expr slope = 1.0f / (high_kink_f - low_kink_f);
        Expr intercept = -(low_kink_f / (high_kink_f - low_kink_f));

        Expr expansion = 255.0f * (slope * i / 255.0f + intercept) + 0.5f;
        return Halide::select(i <= lowKink(0), 0, i < highKink(0), expansion, 255);
    }

    Expr luma(Func f, Var x, Var y) {
        return 0.30f * f(x, y, 0) + 0.59f * f(x, y, 1) + 0.11f * f(x, y, 2);
    }

    Func build() override {
        Func function("function");
        Var x("x"), y("y"), c("c");

        Func histogram("histogram");
        histogram(x) = cast<uint32_t>(0);
        RDom r(0, input.width(), 0, input.height());
        histogram(input(r.x, r.y, 0)) += cast<uint8_t>(1);
        histogram.compute_root();

        Expr pixel_count = input.width() * input.height();
        Func prob("prob");
        Var i;
        prob(i) = histogram(i) / cast<float>(pixel_count);

        Func cumulProb("cumulProb");
        RDom cprob(0, 255);
        cumulProb(i) = prob(i);
        cumulProb(cprob) += cumulProb(cprob-1);
        cumulProb.compute_root();


        Expr pct_in_range = 100.0f * (cumulProb(SHADOW_DETECT_MAX_INTENSITY) - cumulProb(SHADOW_DETECT_MIN_INTENSITY));
        Expr sh_prob_mu = (cumulProb(SHADOW_DETECT_MIN_INTENSITY) + cumulProb(SHADOW_DETECT_MAX_INTENSITY)) * 0.5f;

        Func sh_intensity_mu("sh_intensity_mu");
        RDom shr(SHADOW_DETECT_MIN_INTENSITY, SHADOW_DETECT_MAX_INTENSITY);
//        sh_intensity_mu(i) = cast<uint8_t>(SHADOW_DETECT_MIN_INTENSITY);
//        sh_intensity_mu(shr) += Halide::select(cumulProb(shr) < sh_prob_mu, cast<uint8_t>(1), cast<uint8_t>(0));
//        sh_intensity_mu.compute_root();
        sh_intensity_mu(i) = cast<uint8_t>(0);
        sh_intensity_mu(0) = Halide::maximum(Halide::select(cumulProb(shr) < sh_prob_mu, cast<uint8_t>(shr), cast<uint8_t>(0)));


//        int sh_intensity_mu = SHADOW_DETECT_MIN_INTENSITY;
//        for ( ; sh_intensity_mu <= SHADOW_DETECT_MAX_INTENSITY; sh_intensity_mu++) {
//            if (cumulProb(sh_intensity_mu) >= sh_prob_mu)
//                break;
//        }

        Expr shadow_trans_effect_size = (((float)SHADOW_DETECT_MAX_INTENSITY) - cast<float>((sh_intensity_mu(0)))) / ((float)SHADOW_DETECT_INTENSITY_RANGE) * SHADOW_AGGRESSIVENESS_MUL;

        Func inputFunc("inputFunc");
        inputFunc(x, y, c) = input(x, y, c);
        Expr lumaInput = luma(inputFunc, x, y);

        Func shadowDetailRemapFunc("shadowDetailRemapFunc");
        shadowDetailRemapFunc(i) = shadowDetailRemap(i, shadow_trans_effect_size);
        shadowDetailRemapFunc.compute_root();
        Expr k = Halide::select(lumaInput > 0.0f, clamp(shadowDetailRemapFunc(cast<uint8_t>(lumaInput)), lumaInput, 255) / lumaInput, 1.0f);
        Func shadowDetail;
        shadowDetail(x, y, c) = cast<uint8_t>(clamp(input(x, y, c) * k, 0.0f, 255.0f));
//        function(x, y, c) = shadowDetail(x, y, c);

//        Expr low_discard_mass = DEFAULT_LOW_DISCARD_MASS;
//        Expr high_discard_mass = DEFAULT_HIGH_DISCARD_MASS;
        Expr low_discard_mass = 0.005f;
        Expr high_discard_mass = 0.995f;
        Func toneExpansionRemapFunc("toneExpansionRemapFunc");
        toneExpansionRemapFunc(i) = toneExpansionRemap(i, low_discard_mass, high_discard_mass, cumulProb);
        toneExpansionRemapFunc.compute_root();
        Expr lumaShadowDetail = luma(shadowDetail, x, y);
        Expr kt = Halide::select(lumaShadowDetail > 0.0f, toneExpansionRemapFunc(cast<uint8_t>(lumaShadowDetail)) / lumaShadowDetail, 1.0f);
        Func toneExpansion;
        toneExpansion(x, y, c) = cast<uint8_t>(clamp(shadowDetail(x, y, c) * kt, 0.0f, 255.0f));
//        function(x, y, c) = toneExpansion(x, y, c);

        Expr compensation_multiplier = Halide::select(low_discard_mass < 0.01f, 1.02f, 1.10f);
        Expr grey = 0.3f*toneExpansion(x,y,0) + 0.59f*toneExpansion(x,y,1) + 0.11f*toneExpansion(x,y,2);
        Expr saturate_float = Halide::lerp(grey, Halide::cast<float>(toneExpansion(x, y, c)), compensation_multiplier );
        Expr saturate = Halide::cast<uint8_t>(Halide::clamp(saturate_float, 0.0f, 255.0f));

        function(x, y, c) = select(c == 3, cast<uint8_t>(255), saturate);

//        function(x, y, c) = Halide::select(pct_in_range > 30.0f, shadowDetail, input(x, y, c));
//        Expr x256 = cast<int>(x / cast<float>(input.width()) * 255.0f);
//        function(x, y, c) = cast<uint8_t>(histogram(x256));
//        function(x, y, c) = cast<uint8_t>(clamp(cumulProb(x256) * 255.0f, 0, 255));

        // Schedule for CPU only (glsl fails on this generator)
        function.vectorize(c, 4).parallel(y);

        set_packed(input);
        set_packed(function.output_buffer());

        return function;
    }

};

Halide::RegisterGenerator<AutoEnhance> register_auto_enhance{"auto_enhance"};

}
