#include <QtCore/QDebug>
#include <QtCore/QMetaObject>
#include <QtCore/QMetaProperty>

#include "halide_common.h"
#include "halide_function.h"
#include "halide_image.h"
#include "halide_transform_worker.h"

HalideFunction::HalideFunction(QObject *parent) :
    QObject(parent),
    m_enabled(true),
    m_allPropertiesValid(false),
    m_dirty(true)
{
}

QString HalideFunction::name() const
{
    return m_name;
}

void HalideFunction::setName(QString generatorName)
{
    if (generatorName != m_name) {
        m_name = generatorName;
        matchPropertiesToArguments(newGenerator()->get_filter_arguments());
        Q_EMIT nameChanged();
    }
}

bool HalideFunction::enabled() const
{
    return m_enabled;
}

void HalideFunction::setEnabled(bool enabled)
{
    if (enabled != m_enabled) {
        m_enabled = enabled;
        Q_EMIT enabledChanged();
        setDirty(true);
    }
}

QString halideTypeName(Halide::Type type)
{
    QString name;

    if (type.is_vector()) {
        name += "vector of ";
    }

    if (type.is_bool()) {
        name += "bool";
    } else if (type.is_handle()) {
        name += "handle (void *)";
    } else {
        if (type.is_float()) {
            name += "float";
        } else if (type.is_int()) {
            name += "int";
        } else if (type.is_uint()) {
            name += "unsigned int";
        }
        name += QString(" (%1 bits)").arg(type.bits);
    }

    return name;
}

QString halideTypeToQmlType(Halide::Type type, bool& possibleOverflow)
{
    QString qmlType;
    possibleOverflow = false;

    if (type.is_vector()) {
        qmlType = "N/A";
    }

    if (type.is_bool()) {
        qmlType = "bool";
    } else if (type.is_handle()) {
        qmlType = "N/A";
    } else if (type.is_float()) {
        qmlType = "real";
        possibleOverflow = type.bits < (int)sizeof(double);
    } else if (type.is_int() || type.is_uint()) {
        qmlType = "int";
        possibleOverflow = type.bits < (int)sizeof(int);
    }

    return qmlType;
}

Halide::Type qMetaTypeToHalideType(int qMetaType)
{
    switch (qMetaType) {
    case QMetaType::Bool:
        return Halide::type_of<bool>();
        break;
    case QMetaType::Int:
        return Halide::type_of<int>();
        break;
    case QMetaType::SChar:
    case QMetaType::Char:
        return Halide::type_of<int8_t>();
        break;
    case QMetaType::Short:
        return Halide::type_of<short>();
        break;
    case QMetaType::Long:
        return Halide::type_of<long>();
        break;
    case QMetaType::LongLong:
        return Halide::type_of<int64_t>();
        break;
    case QMetaType::UInt:
        return Halide::type_of<unsigned int>();
        break;
    case QMetaType::UChar:
        return Halide::type_of<unsigned char>();
        break;
    case QMetaType::UShort:
        return Halide::type_of<unsigned short>();
        break;
    case QMetaType::ULong:
        return Halide::type_of<unsigned long>();
        break;
    case QMetaType::ULongLong:
        return Halide::type_of<uint64_t>();
        break;
    case QMetaType::Float:
        return Halide::type_of<float>();
        break;
    case QMetaType::Double:
        return Halide::type_of<double>();
        break;
    default:
        Halide::Type t;
        t.bits = 0;
        t.width = 0;
        return t;
        break;
    }
}

QString halideArgumentToQmlType(const Halide::Argument& argument, bool& possibleOverflow)
{
    if (argument.is_buffer) {
        possibleOverflow = false;
        return "HalideImage";
    } else {
        return halideTypeToQmlType(argument.type, possibleOverflow);
    }
}

bool isHalideArgumentCompatible(const Halide::Argument& argument, int qMetaType)
{
    if (argument.is_buffer) {
        return qMetaType == qMetaTypeId<HalideImage*>();
    } else {
        if (argument.type.code == Halide::Type::TypeCode::UInt && qMetaType == QMetaType::Int) {
            // FIXME: this is a workaround to allow QML int to represent
            // generators' unsigned int parameters; necessary due to the lack of
            // unsigned int in QML
            return true;
        } else {
            return qMetaTypeToHalideType(qMetaType).can_represent(argument.type);
        }
    }
}

QString halideArgumentTypeName(const Halide::Argument& argument)
{
    if (argument.is_buffer) {
        return "ImageParam";
    } else {
        return halideTypeName(argument.type);
    }
}


void HalideFunction::matchPropertiesToArguments(const std::vector<Halide::Argument>& arguments)
{
    m_signalMappers.clear();
    m_propertiesValid.clear();

    const QMetaObject* meta = this->metaObject();
    Q_FOREACH(Halide::Argument argument, arguments) {
        int propertyIndex = meta->indexOfProperty(argument.name.c_str());
        QString propertyName = QString::fromStdString(argument.name);
        m_propertiesValid[propertyName] = false;
        bool possibleOverflow;
        QString expectedQmlType = halideArgumentToQmlType(argument, possibleOverflow);
        QString argumentTypeName = halideArgumentTypeName(argument);
        if (propertyIndex == -1) {
            qWarning().noquote() << QString("HalideFunction '%1' is missing property '%2' of type %3.")
                                           .arg(m_name)
                                           .arg(propertyName)
                                           .arg(expectedQmlType);
        } else {
            int propertyType = meta->property(propertyIndex).userType();
            if (isHalideArgumentCompatible(argument, propertyType)) {
                m_propertiesValid[propertyName] = isPropertyValid(propertyName);
                connectToPropertyChange(propertyName);
            } else {
                QString propertyTypeName = QMetaType::typeName(propertyType);
                qWarning().noquote() << QString("HalideFunction '%1': property '%2' is of type %3 incompatible "
                                                "with generator parameter of Halide type %4. Type should be %5.")
                                               .arg(m_name)
                                               .arg(propertyName)
                                               .arg(propertyTypeName)
                                               .arg(argumentTypeName)
                                               .arg(expectedQmlType);
            }
        }
        if (possibleOverflow) {
            qWarning() << QString("Conversion from QML %1 to %2 could overflow").arg(expectedQmlType).arg(argumentTypeName);
        }
    }
    updateAllPropertiesValid();
}

void HalideFunction::connectToPropertyChange(QString propertyName)
{
    QSharedPointer<QSignalMapper> mapper(new QSignalMapper);
    m_signalMappers.append(mapper);
    mapper->setMapping(this, propertyName);
    connect(mapper.data(), SIGNAL(mapped(QString)), this, SLOT(onPropertyChanged(QString)));

    const QMetaObject* meta = this->metaObject();
    int propertyIndex = meta->indexOfProperty(propertyName.toUtf8().constData());
    QMetaProperty prop = meta->property(propertyIndex);
    QMetaMethod propertyChangedSignal = prop.notifySignal();
    int mapIndex = mapper->metaObject()->indexOfSlot("map()");
    QMetaMethod mapSlot = mapper->metaObject()->method(mapIndex);
    connect(this, propertyChangedSignal, mapper.data(), mapSlot);
}

void HalideFunction::onPropertyChanged(QString propertyName)
{
    QVariant value = this->property(propertyName.toUtf8().constData());
    HalideImage* halideImage = value.value<HalideImage*>();
    if (halideImage) {
        monitorImage(halideImage, propertyName);
    }
    m_propertiesValid[propertyName] = isPropertyValid(propertyName);
    updateAllPropertiesValid();
    setDirty(true);
}

void HalideFunction::updateAllPropertiesValid()
{
    QHash<QString, bool> propertiesValid(m_propertiesValid);
    m_allPropertiesValid = !propertiesValid.values().contains(false);
}

bool HalideFunction::isPropertyValid(QString propertyName)
{
    QVariant value = this->property(propertyName.toUtf8().constData());
    if (value.canConvert<HalideImage*>()) {
        HalideImage* halideImage = value.value<HalideImage*>();
        if (halideImage) {
            return halideImage->data()->buffer.defined();
        } else {
            if (propertyName == "input") {
                // Case where no HalideImage was set happens when chaining HalideFunctions
                return true;
            } else {
                return false;
            }
        }
    } else {
        return true;
    }
}

void HalideFunction::onHalideImageChanged()
{
    HalideImage* halideImage = qobject_cast<HalideImage*>(sender());
    QList<QString> propertyNames = m_halideImages.keys(halideImage);
    Q_FOREACH(QString propertyName, propertyNames) {
        m_propertiesValid[propertyName] = isPropertyValid(propertyName);
    }
    updateAllPropertiesValid();
    setDirty(true);
}

void HalideFunction::monitorImage(HalideImage* halideImage, QString propertyName)
{
    HalideImage* previousHalideImage = m_halideImages.value(propertyName, 0);

    if (halideImage != previousHalideImage) {
        m_halideImages[propertyName] = halideImage;
        if (previousHalideImage && m_halideImages.keys(previousHalideImage).isEmpty()) {
            QObject::disconnect(previousHalideImage, SIGNAL(dataChanged()), this, SLOT(onHalideImageChanged()));
        }
        if (m_halideImages.keys(halideImage).length() == 1) {
            QObject::connect(halideImage, SIGNAL(dataChanged()), this, SLOT(onHalideImageChanged()));
        }
    }
}

bool HalideFunction::allPropertiesValid() const
{
    return m_allPropertiesValid;
}

bool HalideFunction::dirty() const
{
    return m_dirty;
}

void HalideFunction::setDirty(bool dirty)
{
    if (dirty != m_dirty) {
        m_dirty = dirty;
        Q_EMIT dirtyChanged();
    }
}

std::unique_ptr<Halide::Internal::GeneratorBase> HalideFunction::newGenerator() const
{
    Halide::Internal::GeneratorParamValues params;
    return Halide::Internal::GeneratorRegistry::create(m_name.toStdString(), params);
}

void HalideFunction::prepareInputData(QSharedPointer<HalideInputData> inputData) const
{
    // set values to parameters
    unsigned int i = 0;
    Q_FOREACH(Halide::Internal::Parameter parameter, inputData->generator->get_filter_parameters()) {
        QString name = QString::fromStdString(parameter.name());
        QVariant value = this->property(name.toUtf8().constData());
        inputData->setParameter(parameter, i, value);
        i++;
    }
}
