#include "halide_function_list_model.h"

HalideFunctionListModel::HalideFunctionListModel(QObject* parent) : QAbstractListModel(parent)
{
}

void HalideFunctionListModel::clear()
{
    beginRemoveRows(QModelIndex(), 0, m_list.size() - 1);
    m_list.clear();
    endRemoveRows();
    Q_EMIT countChanged();
}

int HalideFunctionListModel::count()
{
    return m_list.size();
}

HalideFunction* HalideFunctionListModel::get(int row)
{
    return data(index(row)).value<HalideFunction*>();
}

void HalideFunctionListModel::append(HalideFunction* function)
{
    insert(m_list.size(), function);
}

void HalideFunctionListModel::append(const QList<HalideFunction*>& functions)
{
    QList<HalideFunction*> nonNullFunctions;
    Q_FOREACH(HalideFunction* function, functions) {
        if (!function) {
            qWarning() << "HalideFunctionListModel::insert: tried to insert a null function";
        } else {
            nonNullFunctions.append(function);
        }
    }
    if (nonNullFunctions.isEmpty()) {
        return;
    }

    int index = m_list.size();
    beginInsertRows(QModelIndex(), index, index + nonNullFunctions.count() - 1);
    m_list.append(nonNullFunctions);
    endInsertRows();
    Q_EMIT countChanged();
}

void HalideFunctionListModel::insert(int index, HalideFunction* function)
{
    if (!function) {
        qWarning() << "HalideFunctionListModel::insert: tried to insert a null function";
        return;
    }
    if (index < 0 || index > m_list.size()) {
        qWarning() << "HalideFunctionListModel::insert: index passed is out of bounds.";
    }
    index = qMax(0, qMin(index, m_list.size()));
    beginInsertRows(QModelIndex(), index, index);
    m_list.insert(index, function);
    endInsertRows();
    Q_EMIT countChanged();
}

bool HalideFunctionListModel::remove(int index)
{
    if (index < 0 || index >= m_list.size()) {
        return false;
    }

    beginRemoveRows(QModelIndex(), index, index);
    m_list.removeAt(index);
    endRemoveRows();
    Q_EMIT countChanged();

    return true;
}

bool HalideFunctionListModel::remove(HalideFunction* function)
{
    return remove(m_list.indexOf(function));
}

bool HalideFunctionListModel::move(int from, int to)
{
    if (from < 0 || from >= m_list.size() || to < 0 || to > m_list.size()) {
        return false;
    }

    beginMoveRows(QModelIndex(), from, from, QModelIndex(), to + (to > from ? 1 : 0));
    m_list.move(from, to);
    endMoveRows();

    Q_FOREACH(HalideFunction* function, m_list.mid(qMin(from, to))) {
        function->setDirty(true);
    }

    return true;
}

QList<HalideFunction*> HalideFunctionListModel::toList()
{
    return m_list;
}

QHash<int, QByteArray> HalideFunctionListModel::roleNames() const
{
    QHash<int, QByteArray> roles;
    roles[ModelData] = "modelData";
    return roles;
}

int HalideFunctionListModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return m_list.size();
}

QVariant HalideFunctionListModel::data(const QModelIndex &index, int role) const
{
    Q_UNUSED(role)
    if (!index.isValid()) {
        return QVariant();
    }

    if (index.row() < 0 || index.row() >= m_list.size()) {
        return QVariant();
    }

    HalideFunction* function = m_list.at(index.row());
    return QVariant::fromValue(function);
}
