#include "halide_image.h"
#include "halide_common.h"
#include <QtCore/QDebug>

HalideImage::HalideImage(QObject *parent) :
    QObject(parent),
    m_componentCompleted(false),
    m_source(""),
    m_status(HalideImage::Null),
    m_data(new HalideImageData),
    m_reloadPending(false)
{
    connect(&m_reloadWatcher, SIGNAL(finished()), this, SLOT(onReloadSourceFinished()));
}

HalideImage::~HalideImage()
{
}

void HalideImage::classBegin()
{
}

void HalideImage::componentComplete()
{
    // FIXME: this is called before m_sourceSize has its final value leading
    // to many pictures reload
    m_componentCompleted = true;
    scheduleReloadSource();
}

QUrl HalideImage::source() const
{
    return m_source;
}

QSize HalideImage::sourceSize() const
{
    return m_sourceSize;
}

QSize HalideImage::size() const
{
    if (m_data) {
        return m_data->size;
    } else {
        return QSize();
    }
}

HalideImage::Status HalideImage::status() const
{
    return m_status;
}

void HalideImage::setStatus(HalideImage::Status status)
{
    if (status != m_status) {
        m_status = status;
        Q_EMIT statusChanged(status);
    }
}

void HalideImage::setSource(QUrl source)
{
    if (source != m_source) {
        m_source = source;
        scheduleReloadSource();
        Q_EMIT sourceChanged();
    }
}

void HalideImage::setSourceSize(QSize sourceSize)
{
    if (sourceSize != m_sourceSize) {
        m_sourceSize = sourceSize;
        scheduleReloadSource();
        Q_EMIT sourceSizeChanged();
    }
}

void HalideImage::setData(QSharedPointer<HalideImageData> data)
{
    Q_ASSERT_X(!data.isNull(), "HalideImage::setData", "cannot set null data");

    if (data.isNull()) {
        qWarning() << "HalideImage::setData received null data. This is forbidden.";
    }
    QSharedPointer<HalideImageData> oldData = m_data;
    if (data != oldData) {
        m_data = data;

        if (!m_source.isEmpty()) {
            m_source.clear();
            Q_EMIT sourceChanged();
        }

        if (data->size != oldData->size) {
            Q_EMIT sizeChanged();
        }
    }
    Q_EMIT dataChanged();
}

QSharedPointer<HalideImageData> HalideImage::data()
{
    return m_data;
}

QSharedPointer<HalideImageData> HalideImage::reloadSource(QUrl source, QSize sourceSize)
{
    QSharedPointer<HalideImageData> newData(new HalideImageData);
    if (source.isValid()) {
        m_reader.setFileName(source.path());
        QSize imageSize = m_reader.size();
        if (!sourceSize.isNull() &&
            (imageSize.width() > sourceSize.width() ||
             imageSize.height() > sourceSize.height())) {
            QSize validRequestedSize = sourceSize;
            if (validRequestedSize.width() == 0) {
                validRequestedSize.setWidth(imageSize.width());
            }
            if (validRequestedSize.height() == 0) {
                validRequestedSize.setHeight(imageSize.height());
            }
            imageSize.scale(validRequestedSize, Qt::KeepAspectRatio);
            m_reader.setScaledSize(imageSize);
        }
        newData->image = m_reader.read();
        newData->isPacked = true;
        newData->size = newData->image.size();
        newData->image = newData->image.convertToFormat(QImage::Format_RGBA8888);
        if (!newData->image.isNull()) {
            newData->buffer = halideBufferFromQImage(newData->image);
        } else {
            newData->buffer = Halide::Buffer();
        }
    }
    return newData;
}

void HalideImage::onReloadSourceFinished()
{
    if (m_reloadPending) {
        m_reloadPending = false;
        scheduleReloadSource();
    } else {
        QFutureWatcher< QSharedPointer<HalideImageData> >* watcher;
        watcher = static_cast< QFutureWatcher< QSharedPointer<HalideImageData> >* >(sender());
        m_data = watcher->result();
        Q_EMIT sizeChanged();
        Q_EMIT dataChanged();

        if (!m_data->image.isNull()) {
            setStatus(HalideImage::Ready);
        } else {
            setStatus(HalideImage::Error);
        }
    }
}

void HalideImage::scheduleReloadSource()
{
    if (!m_componentCompleted) {
        return;
    }

    if (m_reload.isFinished()) {
        if (m_source.isEmpty()) {
            setStatus(HalideImage::Null);
        } else {
            setStatus(HalideImage::Loading);
            m_reload = QtConcurrent::run(this, &HalideImage::reloadSource, m_source, m_sourceSize);
            m_reloadWatcher.setFuture(m_reload);
        }
    } else {
        m_reloadPending = true;
    }
}
