#include "halide_image_renderer.h"
#include <QtQuick/QQuickWindow>

HalideImageRenderer::HalideImageRenderer(QQuickItem *parent) :
    QQuickItem(parent),
    m_textureNeedsUpdate(true),
    m_image(0),
    m_fillMode(HalideImageRenderer::PreserveAspectFit),
    m_horizontalAlignment(HalideImageRenderer::AlignHCenter),
    m_verticalAlignment(HalideImageRenderer::AlignVCenter),
    m_window(0)
{
    setFlag(QQuickItem::ItemHasContents);
    connectToRendering(window());
    connect(this, SIGNAL(windowChanged(QQuickWindow*)),
            this, SLOT(connectToRendering(QQuickWindow*)));

}

HalideImageRenderer::~HalideImageRenderer()
{
}

void HalideImageRenderer::connectToRendering(QQuickWindow* window)
{
    if (m_window) {
        disconnect(m_window, SIGNAL(beforeRendering()), this, SLOT(lockImage()));
        disconnect(m_window, SIGNAL(afterRendering()), this, SLOT(unlockImage()));
    }
    m_window = window;
    if (m_window) {
        connect(m_window, SIGNAL(beforeRendering()), this, SLOT(lockImage()),
                Qt::DirectConnection);
        connect(m_window, SIGNAL(afterRendering()), this, SLOT(unlockImage()),
                Qt::DirectConnection);
    }
}

QSGNode* HalideImageRenderer::updatePaintNode(QSGNode* oldNode, UpdatePaintNodeData* updatePaintNodeData)
{
    Q_UNUSED(updatePaintNodeData)

    if (width() <= 0 || height() <= 0) {
        if (oldNode) {
            QSGTexture* oldTexture = static_cast<QSGSimpleTextureNode*>(oldNode)->texture();
            delete oldNode;
            if (oldTexture) {
                delete oldTexture;
            }
        }
        return NULL;
    } else {
        QSGSimpleTextureNode* node = static_cast<QSGSimpleTextureNode*>(oldNode);
        if (!node) {
            node = new QSGSimpleTextureNode;
            node->setFiltering(QSGTexture::Linear);
            QSGTexture* texture = window()->createTextureFromId(0, QSize());
            node->setTexture(texture);
        }

        if (m_textureNeedsUpdate) {
            QSGTexture* oldTexture = node->texture();
            if (oldTexture) {
                delete oldTexture;
            }

            if (m_image) {
                m_imageData = m_image->data();
            } else {
                m_imageData.clear();
            }
            QSGTexture* texture = createTextureFromImageData(m_imageData);
            if (!texture) {
                texture = window()->createTextureFromId(0, QSize());
            }

            node->setTexture(texture);
            m_textureNeedsUpdate = false;
        }

        updateGeometry(node);
        return node;
    }
}

void HalideImageRenderer::updateGeometry(QSGSimpleTextureNode* node)
{
    QRectF nodeRect;
    if (node->texture()) {
        QSize textureSize = node->texture()->textureSize();
        if (textureSize.isValid() && !textureSize.isEmpty()) {
            QRectF bounds = boundingRect();
            qreal imageRatio = textureSize.width() / (qreal)textureSize.height();
            qreal frameRatio = bounds.width() / bounds.height();
            bool fillRatioCondition = true;
            if (m_fillMode == HalideImageRenderer::PreserveAspectFit) {
                fillRatioCondition = (imageRatio >= frameRatio);
            } else if (m_fillMode == HalideImageRenderer::PreserveAspectCrop) {
                fillRatioCondition = (imageRatio <= frameRatio);
            }
            QSize size = fillRatioCondition ? QSize(bounds.width(), bounds.width() / imageRatio) :
                                              QSize(bounds.height() * imageRatio, bounds.height());
            nodeRect.setSize(size);

            qreal tx = 0.0f;
            switch (m_horizontalAlignment) {
            case AlignLeft:
                tx = 0.0f;
                break;
            case AlignRight:
                tx = bounds.width() - size.width();
                break;
            case AlignHCenter:
                tx = (bounds.width() - size.width()) / 2.0f;
                break;
            }

            qreal ty = 0.0f;
            switch (m_verticalAlignment) {
            case AlignTop:
                ty = 0.0f;
                break;
            case AlignBottom:
                ty = bounds.height() - size.height();
                break;
            case AlignVCenter:
                ty = (bounds.height() - size.height()) / 2.0f;
                break;
            }
            nodeRect.translate(tx, ty);
        } else {
            nodeRect = boundingRect();
        }
    }

    node->setRect(nodeRect);
}

void HalideImageRenderer::lockImage()
{
    // No need to lock if the data is backed by a QImage as it is not used
    // during rendering (pixels have been uploaded to a texture)
    if (m_imageData && m_imageData->fbo) {
        m_imageData->lock.lockForRead();
    }
}

void HalideImageRenderer::unlockImage()
{
    if (m_imageData && m_imageData->fbo) {
        m_imageData->lock.unlock();
    }
}

QSGTexture* HalideImageRenderer::createTextureFromImageData(QSharedPointer<HalideImageData> imageData)
{
    QSGTexture* texture = NULL;
    if (imageData) {
        m_imageData->lock.lockForRead();
        if (imageData->fbo && imageData->fbo->texture() > 0) {
            texture = window()->createTextureFromId(imageData->fbo->texture(), m_image->size());
        } else if (!imageData->image.isNull()) {
            texture = window()->createTextureFromImage(imageData->image);
        }
        m_imageData->lock.unlock();
    }
    return texture;
}

HalideImage* HalideImageRenderer::image() const
{
    return m_image;
}

void HalideImageRenderer::setImage(HalideImage* image)
{
    if (image != m_image) {
        if (m_image != NULL) {
            QObject::disconnect(m_image, SIGNAL(dataChanged()), this, SLOT(scheduleUpdate()));
        }
        m_image = image;
        if (m_image != NULL) {
            QObject::connect(m_image, SIGNAL(dataChanged()), this, SLOT(scheduleUpdate()));
        } else {
            setSize(QSizeF(0, 0));
        }
        scheduleUpdate();
        Q_EMIT imageChanged();
    }
}

HalideImageRenderer::FillMode HalideImageRenderer::fillMode() const
{
    return m_fillMode;
}

void HalideImageRenderer::setFillMode(FillMode fillMode)
{
    if (fillMode != m_fillMode) {
        m_fillMode = fillMode;
        scheduleUpdate();
        Q_EMIT fillModeChanged();
    }
}

HalideImageRenderer::HorizontalAlignment HalideImageRenderer::horizontalAlignment() const
{
    return m_horizontalAlignment;
}

void HalideImageRenderer::setHorizontalAlignment(HorizontalAlignment horizontalAlignment)
{
    if (horizontalAlignment != m_horizontalAlignment) {
        m_horizontalAlignment = horizontalAlignment;
        scheduleUpdate();
        Q_EMIT horizontalAlignmentChanged();
    }
}

HalideImageRenderer::VerticalAlignment HalideImageRenderer::verticalAlignment() const
{
    return m_verticalAlignment;
}

void HalideImageRenderer::setVerticalAlignment(VerticalAlignment verticalAlignment)
{
    if (verticalAlignment != m_verticalAlignment) {
        m_verticalAlignment = verticalAlignment;
        scheduleUpdate();
        Q_EMIT verticalAlignmentChanged();
    }
}

void HalideImageRenderer::scheduleUpdate()
{
    if (m_image) {
        setSize(m_image->size());
    }
    m_textureNeedsUpdate = true;
    QQuickItem::update();
}
