#include "halide_transform.h"
#include "halide_function_list_model.h"
#include "halide_image.h"
#include "halide_common.h"
#include "halide_transform_cpu.h"
#include "halide_transform_opengl.h"
#include <QtCore/QDebug>
#include <QtCore/QStandardPaths>
#include <QtQuick/QQuickItem>
#include <QtConcurrent/QtConcurrent>

HalideTransform::HalideTransform(QObject *parent) :
    QObject(parent),
    m_componentCompleted(false),
    m_functionsModel(new HalideFunctionListModel(this)),
    m_live(false),
    m_worker(0),
    m_input(0),
    m_output(new HalideImage(this)),
    m_window(0),
    m_parent(parent),
    m_updateInProgress(false),
    m_pendingUpdate(false),
    m_pendingRequiresImage(false)
{
    connect(m_functionsModel, SIGNAL(rowsInserted(QModelIndex,int,int)),
            this, SLOT(onFunctionsInserted(QModelIndex,int,int)));
    connect(m_functionsModel, SIGNAL(rowsAboutToBeRemoved(QModelIndex,int,int)),
            this, SLOT(onFunctionsAboutToBeRemoved(QModelIndex,int,int)));
    connect(m_functionsModel, SIGNAL(rowsRemoved(QModelIndex,int,int)),
            this, SLOT(onFunctionsRemoved(QModelIndex,int,int)));

    const Halide::Target &target = Halide::get_jit_target_from_environment();
    if (target.has_feature(Halide::Target::OpenGL))  {
        m_worker = new HalideTransformOpenGL(this);
    } else {
        m_worker = new HalideTransformCpu(this);
    }
    connect(m_worker, SIGNAL(updateRequired()), this, SLOT(scheduleUpdate()), Qt::QueuedConnection);
    connect(m_worker, SIGNAL(computeFinished(QSharedPointer<HalideImageData>)),
            this, SLOT(onWorkedComputeFinished(QSharedPointer<HalideImageData>)), Qt::QueuedConnection);
    setupOpenGLContextMonitoring();
}

HalideTransform::~HalideTransform()
{
    if (m_worker) {
        m_worker->deleteLater();
        m_worker = 0;
    }
    delete m_functionsModel;
}

void HalideTransform::classBegin()
{
    // we know that parent() is not always set during instantiation
    // but is most likely set at this point
    setupOpenGLContextMonitoring();
}

void HalideTransform::componentComplete()
{
    m_componentCompleted = true;

    if (m_live) {
        scheduleUpdate();
    }
}

bool HalideTransform::live() const
{
    return m_live;
}

void HalideTransform::setLive(bool live)
{
    if (live != m_live) {
        m_live = live;
        bool anyDirty = false;

        Q_FOREACH(HalideFunction* function, m_functionsModel->toList()) {
            if (m_live) {
                connect(function, SIGNAL(dirtyChanged()), this, SLOT(functionMayNeedUpdate()));
                anyDirty = true;
            } else {
                disconnect(function, SIGNAL(dirtyChanged()), this, SLOT(functionMayNeedUpdate()));
            }
        }

        if (anyDirty) {
            scheduleUpdate();
        }

        Q_EMIT liveChanged();
    }
}

bool HalideTransform::updating() const
{
    return m_updateTimer.isValid();
}

HalideImage* HalideTransform::input() const
{
    return m_input;
}

void HalideTransform::setInput(HalideImage* input)
{
    if (input != m_input) {
        if (m_input) {
            QObject::disconnect(m_input, SIGNAL(dataChanged()), this, SLOT(onInputChanged()));
        }

        m_input = input;

        if (m_input) {
            QObject::connect(m_input, SIGNAL(dataChanged()), this, SLOT(onInputChanged()));
            if (m_functionsModel->count() == 0) {
                if (m_input) {
                    m_output->setData(m_input->data());
                }
            }
        }

        Q_EMIT inputChanged();
        onInputChanged();
    }
}

void HalideTransform::onInputChanged()
{
    if (m_functionsModel->count() == 0) {
        if (m_input) {
            m_output->setData(m_input->data());
        }
    } else {
        m_functionsModel->get(0)->setDirty(true);
    }
}

void HalideTransform::onFunctionsAboutToBeRemoved(const QModelIndex & parent, int first, int last)
{
    Q_UNUSED(parent)

    QList<HalideFunction*> functions = m_functionsModel->toList().mid(first, last - first + 1);
    Q_FOREACH(HalideFunction* function, functions) {
        disconnect(function, SIGNAL(dirtyChanged()), this, SLOT(functionMayNeedUpdate()));
    }
}

void HalideTransform::onFunctionsRemoved(const QModelIndex & parent, int first, int last)
{
    Q_UNUSED(parent)

    m_worker->pruneCache(m_functionsModel->toList());

    if (m_live) {
        scheduleUpdate();
    }
}

void HalideTransform::onFunctionsInserted(const QModelIndex & parent, int first, int last)
{
    Q_UNUSED(parent)

    QList<HalideFunction*> functions = m_functionsModel->toList().mid(first, last - first + 1);
    Q_FOREACH(HalideFunction* function, functions) {
        function->setDirty(true);
        if (m_live) {
            connect(function, SIGNAL(dirtyChanged()), this, SLOT(functionMayNeedUpdate()));
        }
    }

    if (m_live) {
        scheduleUpdate();
    }
}

HalideImage* HalideTransform::output() const
{
    return m_output;
}

void HalideTransform::functionMayNeedUpdate()
{
    HalideFunction* senderFunction = qobject_cast<HalideFunction*>(sender());
    if (senderFunction && senderFunction->dirty()) {
        doUpdate(false);
    }
}

void HalideTransform::scheduleUpdate()
{
    doUpdate(false);
}

void HalideTransform::doUpdate(bool outputImageRequired)
{
    if (!m_componentCompleted) {
        return;
    }

    if (m_functionsModel->count() == 0) {
        if (m_input) {
            m_output->setData(m_input->data());
        }
        return;
    }

    Q_FOREACH(HalideFunction* function, m_functionsModel->toList()) {
        if (!function->allPropertiesValid()) {
            function->setDirty(false);
            qDebug() << "doUpdate(): missing parameters for Halide function.";
            return;
        }
    }

    // if first function has an "input" property, verify that
    // either it is set to a valid image or HalideTransform::input is
    QVariant inputProperty = m_functionsModel->get(0)->property("input");
    bool hasInputProperty = inputProperty.isValid();
    if (hasInputProperty) {
        HalideImage* value = inputProperty.value<HalideImage*>();
        if (!value || !value->data()->buffer.defined()) {
            if (!m_input || !m_input->data()->buffer.defined()) {
                m_functionsModel->get(0)->setDirty(false);
                qDebug() << "HalideTransform's input does not have a valid image.";
                return;
            }
        }
    }

    if (!m_worker || !m_worker->isReady()) {
        return;
    }

    if (!m_updateInProgress) {
        m_updateFunctionNames.clear();
        Q_FOREACH(HalideFunction* function, m_functionsModel->toList()) {
            m_updateFunctionNames.append(function->name());
        }
        m_updateInProgress = true;
        m_updateTimer.start();
        Q_EMIT updatingChanged();
        m_worker->compute(m_input, m_functionsModel->toList(), outputImageRequired);
    } else {
        m_pendingUpdate = true;
        m_pendingRequiresImage = outputImageRequired;
    }
}

void HalideTransform::onWorkedComputeFinished(QSharedPointer<HalideImageData> result)
{
    m_output->setData(result);

    qDebug() << QString("\"%1\" applied in %2 ms (%3x%4)")
                .arg(m_updateFunctionNames.join(" + "))
                .arg(m_updateTimer.elapsed())
                .arg(m_output->size().width())
                .arg(m_output->size().height()).toLocal8Bit().constData();
    m_updateTimer.invalidate();
    Q_EMIT updatingChanged();
    m_updateInProgress = false;

    if (!m_pendingSaveFiles.isEmpty()) {
        QtConcurrent::run(this, &HalideTransform::saveImageToFiles, m_output->data()->image, m_pendingSaveFiles);
        m_pendingSaveFiles.clear();
    }

    if (m_pendingUpdate) {
        m_pendingUpdate = false;
        m_pendingRequiresImage = false;
        doUpdate(m_pendingRequiresImage);
    }
}

bool HalideTransform::saveToFile(QString fileName)
{
    QString directory = QStandardPaths::writableLocation(QStandardPaths::PicturesLocation);
    QString filePath = directory + "/" + fileName;

    if (fileName.isEmpty()) {
        qWarning() << "Empty file name to passed to HalideTransform::saveToFile";
        return false;
    }

    if (m_output->data()->image.isNull() || !m_live) {
        m_pendingSaveFiles.append(filePath);
        doUpdate(true);
        return true;
    } else {
        QtConcurrent::run(this, &HalideTransform::saveImageToFiles, m_output->data()->image, QStringList(filePath));
        return true;
    }
}

void HalideTransform::saveImageToFiles(QImage image, QStringList filePaths)
{
    bool success;
    Q_FOREACH(QString filePath, filePaths) {
        success = image.save(filePath);
        if (!success) {
            qWarning() << "Failed to save image to file:" << filePath;
        }
    }
}

QVariant HalideTransform::functions() const
{
    return QVariant::fromValue(m_functionsModel);
}

void HalideTransform::setFunctions(QVariant functions)
{
    if (functions.userType() == qMetaTypeId<QJSValue>()) {
        functions = functions.value<QJSValue>().toVariant();
    }

    /* At this point functions can be either:
     * - QVariant(HalideFunction*)
     * - QVariant(QVariantList, (QVariant(HalideFunction*), ... ) )
     *
     * Any other type will be rejected.
     *
     * Additionally QVariant(HalideFunction*) is only accepted during
     * HalideTransform's initialization in order to support the syntax:
     *
     *     HalideTransform {
     *         HalideFunction {
     *         }
     *         HalideFunction {
     *         }
     *     }
     */

    if (functions.type() == QVariant::List) {
        QList<HalideFunction*> functionsList;
        QSequentialIterable iterable = functions.value<QSequentialIterable>();
        Q_FOREACH(const QVariant &functionVariant, iterable) {
            HalideFunction* function = functionVariant.value<HalideFunction*>();
            functionsList.append(function);
        }
        m_functionsModel->clear();
        m_functionsModel->append(functionsList);
    } else if (!m_componentCompleted && functions.canConvert<HalideFunction*>()) {
        HalideFunction* function = functions.value<HalideFunction*>();
        m_functionsModel->append(function);
    } else {
        qWarning() << "HalideTransform::setFunctions received an argument of the wrong type."
                      " Expected a QVariant(QVariantList, (QVariant(HalideFunction*), ... ) )";
    }
}

void HalideTransform::setupOpenGLContextMonitoring()
{
    if (m_parent == parent()) {
        return;
    }

    if (m_parent) {
        QQuickItem* previousParentItem = qobject_cast<QQuickItem*>(m_parent);
        disconnect(previousParentItem, SIGNAL(windowChanged(QQuickWindow*)),
                   this, SLOT(monitorOpenGLContext(QQuickWindow*)));
        monitorOpenGLContext(0);
    }

    m_parent = parent();

    QQuickItem* parentItem = qobject_cast<QQuickItem*>(m_parent);
    if (parentItem) {
        QQuickWindow* window = parentItem->window();
        monitorOpenGLContext(window);
        connect(parentItem, SIGNAL(windowChanged(QQuickWindow*)),
                this, SLOT(monitorOpenGLContext(QQuickWindow*)));
    } else {
        qWarning() << "HalideTransform's parent has to be a QQuickItem.";
    }
}

void HalideTransform::monitorOpenGLContext(QQuickWindow* window)
{
    if (m_window) {
        disconnect(m_window, 0, this, 0);
        Q_EMIT openglContextDestroyed();
    }

    m_window = window;
    if (m_window != NULL) {
        connect(m_window, SIGNAL(beforeSynchronizing()),
                this, SLOT(emitOpenglContextCreated()),
                Qt::DirectConnection);
        connect(m_window, SIGNAL(sceneGraphInitialized()),
                this, SLOT(emitOpenglContextCreated()),
                Qt::DirectConnection);
        connect(m_window, SIGNAL(sceneGraphInvalidated()),
                this, SLOT(emitOpenglContextDestroyed()),
                Qt::DirectConnection);
    }
}

void HalideTransform::emitOpenglContextCreated()
{
    disconnect(m_window, SIGNAL(sceneGraphInitialized()),
               this, SLOT(emitOpenglContextCreated()));
    disconnect(m_window, SIGNAL(beforeSynchronizing()),
               this, SLOT(emitOpenglContextCreated()));
    Q_EMIT openglContextCreated(m_window->openglContext());
}

void HalideTransform::emitOpenglContextDestroyed()
{
    connect(m_window, SIGNAL(beforeSynchronizing()),
            this, SLOT(emitOpenglContextCreated()),
            Qt::DirectConnection);
    connect(m_window, SIGNAL(sceneGraphInitialized()),
            this, SLOT(emitOpenglContextCreated()),
            Qt::DirectConnection);
    Q_EMIT openglContextDestroyed();
}
