#include "halide_transform_opengl.h"
#include "halide_transform.h"
#include "halide_common.h"
#include <QtCore/QDebug>

HalideTransformOpenGL::HalideTransformOpenGL(HalideTransform* transform) :
    HalideTransformWorker(transform),
    m_context(0)
{
    // FIXME: make to work in AOT as well (only works with JIT now)
    const Halide::Target &target = Halide::get_jit_target_from_environment();
    if (!target.has_feature(Halide::Target::OpenGL))  {
        qWarning() << "HalideTransformOpenGL requires running with an OpenGL target. Set HL_JIT_TARGET=host-opengl.";
    }
    connect(m_transform, SIGNAL(openglContextCreated(QOpenGLContext*)),
            this, SLOT(createOpenGLContext(QOpenGLContext*)), Qt::DirectConnection);
    connect(m_transform, SIGNAL(openglContextDestroyed()),
            this, SLOT(destroyOpenGLContext()), Qt::QueuedConnection);
}

HalideTransformOpenGL::~HalideTransformOpenGL()
{
    destroyOpenGLContext();
}

bool HalideTransformOpenGL::isReady()
{
    return m_context && m_context->isValid();
}

void HalideTransformOpenGL::preRender()
{
    m_preRenderContext = QOpenGLContext::currentContext();
    if (m_preRenderContext) {
        m_preRenderSurface = m_preRenderContext->surface();
    }
    m_context->makeCurrent(&m_surface);
}

void HalideTransformOpenGL::postRender()
{
    m_context->doneCurrent();
    if (m_preRenderContext) {
        m_preRenderContext->makeCurrent(m_preRenderSurface);
    }
}

QSharedPointer<HalideImageData> HalideTransformOpenGL::prepareOutputData(QSharedPointer<HalideImageData> oldOutputData, QSize outputSize)
{
    if (outputSize != oldOutputData->size) {
        QSharedPointer<HalideImageData> outputData(new HalideImageData);
        // only reallocate a buffer if necessary
        QSharedPointer<QOpenGLFramebufferObject> outputFbo;
        QOpenGLFramebufferObjectFormat format;
        Halide::Buffer output_buffer;
        outputFbo.reset(new QOpenGLFramebufferObject(outputSize, format));
        outputFbo->bind();
        output_buffer = halideBufferFromTexture(outputFbo->texture(), outputSize);
        outputData->isPacked = true;
        outputData->fbo = outputFbo;
        outputData->buffer = output_buffer;
        outputData->size = outputSize;
        return outputData;
    } else {
        oldOutputData->fbo->bind();
        oldOutputData->buffer.raw_buffer()->dev = oldOutputData->fbo->texture();
        oldOutputData->size = outputSize;
        return oldOutputData;
    }
}


void HalideTransformOpenGL::createOpenGLContext(QOpenGLContext* parentContext)
{
    m_context = new QOpenGLContext;

    QSurfaceFormat format;
    QOpenGLContext* currentContext = QOpenGLContext::currentContext();
    QSurface* currentContextSurface = NULL;
    if (currentContext) {
        currentContextSurface = currentContext->surface();
        // Some GL implementations require that the currently bound context is
        // made non-current before we set up sharing
        currentContext->doneCurrent();
    }
    if (parentContext) {
        format = parentContext->format();
        m_context->setShareContext(parentContext);
    } else {
        format.setDepthBufferSize( 24 );
    }

    m_context->setFormat(format);
    m_context->create();
    m_context->moveToThread(m_thread);

    // surface.create() should only be called on main (GUI) thread
    QMetaObject::invokeMethod(this, "createSurface");

    if (currentContext) {
        currentContext->makeCurrent(currentContextSurface);
    }
    if (m_context->isValid()) {
        Q_EMIT updateRequired();
    }
}

void HalideTransformOpenGL::destroyOpenGLContext()
{
    if (m_context) {
        m_context->makeCurrent(&m_surface);
        m_cache.clear();
        m_context->doneCurrent();

        delete m_context;
        m_context = 0;

        if (m_surface.isValid()) {
            /* From Qt's documentation:
             * "Due to the fact that QOffscreenSurface is backed by a QWindow on some platforms,
             * cross-platform applications must ensure that create() is only called on the main
             * (GUI) thread. The QOffscreenSurface is then safe to be used with makeCurrent()
             * on other threads, but the initialization and destruction must always happen on
             * the main (GUI) thread."
             */
            m_surface.deleteLater();
        }
    }
    m_thread->quit();
}

void HalideTransformOpenGL::createSurface()
{
    m_surface.setFormat(m_context->format());
    m_surface.create();
    this->moveToThread(m_thread);
    m_thread->start();
}
