#include "halide_transform_worker.h"
#include "halide_common.h"
#include "halide_function.h"
#include <QtCore/QVariant>
#include <QtCore/QDebug>
#include <QtCore/QStringList>
#include <dlfcn.h>

QMutex HalideTransformWorker::s_jitLock;

HalideTransformWorker::HalideTransformWorker(HalideTransform* transform) :
    QObject(0),
    m_transform(transform),
    m_thread(new QThread),
    m_outputImage(new HalideImageData)
{
}

HalideTransformWorker::~HalideTransformWorker()
{
    m_thread->quit();
    connect(m_thread, SIGNAL(finished()), m_thread, SLOT(deleteLater()));
}

void HalideTransformWorker::compute(HalideImage* inputImage, QList<HalideFunction*> functions, bool qimageRequired)
{
    QSharedPointer<HalideImageData> lastOutput;
    QList<HalideFunction*> computedFunctions = functionsToRecompute(functions, lastOutput);

    if (computedFunctions.isEmpty()) {
        if (lastOutput) {
            m_outputImage = lastOutput;
        } else if (inputImage) {
            m_outputImage = inputImage->data();
        } else {
            m_outputImage.reset(new HalideImageData);
        }
        Q_EMIT computeFinished(m_outputImage);
    } else {
        QSharedPointer<HalideImageData> inputData;
        if (lastOutput) {
            inputData = lastOutput;
        } else if (inputImage) {
            inputData = inputImage->data();
        }

        QList< QSharedPointer<HalideFunctionData> > functionsData = prepareFunctionsData(computedFunctions);
        functionsData.first()->input->setBufferParameter("input", inputData);
        QMetaObject::invokeMethod(this, "realizeChainFunctions", Qt::QueuedConnection,
                                  Q_ARG(QList< QSharedPointer<HalideFunctionData> >, functionsData),
                                  Q_ARG(bool, qimageRequired));
    }
}

void HalideTransformWorker::pruneCache(QList<HalideFunction*> validFunctions)
{
    QHash<HalideFunction*, QSharedPointer<HalideFunctionData> > newCache;
    QHash<HalideFunction*, QSharedPointer<HalideFunctionData> >::iterator i;
    Q_FOREACH(HalideFunction* function, validFunctions) {
        i = m_cache.find(function);
        if (i != m_cache.end()) {
            newCache[function] = i.value();
        }
    }

    m_cache = newCache;
}

QList<HalideFunction*> HalideTransformWorker::functionsToRecompute(QList<HalideFunction*> functions, QSharedPointer<HalideImageData>& lastOutput) const
{
    QList<HalideFunction*> recomputedFunctions;
    QStringList recomputedFunctionNames;
    QStringList notRecomputedFunctionsNames;
    bool recomputeRemaining = false;

    Q_FOREACH(HalideFunction* function, functions) {
        if (!recomputeRemaining || function->dirty() ||
            (m_cache.contains(function) && !m_cache[function]->outputContentsValid)) {
            recomputeRemaining = true;
        }

        function->setDirty(false);

        if (function->enabled()) {
            if (recomputeRemaining) {
                recomputedFunctions.append(function);
                recomputedFunctionNames.append(function->name());
                if (function != functions.first()) {
                    function->setProperty("input", QVariant());
                }
            } else {
                lastOutput = m_cache[function]->output;
                notRecomputedFunctionsNames.append(function->name());
            }
        }
    }

    if (recomputedFunctionNames.isEmpty()) {
        qDebug() << "Transform was already computed, using cached result.";
    } else if (notRecomputedFunctionsNames.isEmpty()) {
        qDebug() << QString("Transform recomputed %1").arg(recomputedFunctionNames.join("+"));
    } else {
        qDebug() << QString("Transform used %1 from cache and recomputed %2")
                    .arg(notRecomputedFunctionsNames.join("+")).arg(recomputedFunctionNames.join("+"));
    }

    return recomputedFunctions;
}

QList< QSharedPointer<HalideFunctionData> > HalideTransformWorker::prepareFunctionsData(QList<HalideFunction*> functions)
{
    QList< QSharedPointer<HalideFunctionData> > functionsData;

    Q_FOREACH(HalideFunction* function, functions) {
        QSharedPointer<HalideFunctionData> functionData;
        if (!m_cache.contains(function)) {
            functionData.reset(new HalideFunctionData);
            functionData->input.reset(new HalideInputData(function->name()));
            functionData->output.reset(new HalideImageData);
            m_cache[function] = functionData;
        } else {
            functionData = m_cache[function];
        }
        functionData->outputContentsValid = false;
        function->prepareInputData(functionData->input);
        functionsData.append(functionData);
    }

    return functionsData;
}

void HalideTransformWorker::realizeChainFunctions(QList< QSharedPointer<HalideFunctionData> > functionsData, bool qimageRequired)
{
    preRealizeChainFunctions();

    QSharedPointer<HalideImageData> inputData;
    inputData = functionsData.first()->input->getBufferParameter("input");

    Q_FOREACH(QSharedPointer<HalideFunctionData> functionData, functionsData) {
        functionData->input->setBufferParameter("input", inputData);
        QRect outputGeometry = getOutputGeometry(functionData->input->generator.get());

        functionData->output = prepareOutputData(functionData->output, outputGeometry.size());
        functionData->output->lock.lockForWrite();
        functionData->output->buffer.set_min(outputGeometry.x(), outputGeometry.y());

        Q_FOREACH(QSharedPointer<HalideImageData> imageData, functionData->input->imageDataCache) {
            imageData->lock.lockForRead();
        }
        realizeFunction(functionData);
        Q_FOREACH(QSharedPointer<HalideImageData> imageData, functionData->input->imageDataCache) {
            imageData->lock.unlock();
        }
        functionData->output->lock.unlock();
        functionData->outputContentsValid = true;

        functionData->output->buffer.set_min(0, 0);
        inputData = functionData->output;
    }

    QSharedPointer<HalideImageData> lastOutput = functionsData.last()->output;

    if (qimageRequired && lastOutput->fbo) {
        lastOutput->lock.lockForWrite();
#if (QT_VERSION >= QT_VERSION_CHECK(5, 4, 0))
        lastOutput->image = lastOutput->fbo->toImage(false);
#else
        lastOutput->image = lastOutput->fbo->toImage().mirrored();
#endif
        lastOutput->lock.unlock();
    }

    // FIXME: only useful if output is going to be used for reading while being
    // written on again by this worker
    qSwap(m_outputImage, functionsData.last()->output);
    functionsData.last()->outputContentsValid = false;

    postRealizeChainFunctions();

    Q_EMIT computeFinished(m_outputImage);
}

void HalideTransformWorker::realizeFunction(QSharedPointer<HalideFunctionData> functionData)
{
    QSharedPointer<HalideInputData> inputData = functionData->input;

    if (inputData->hasExternFunction) {
        void** arguments = inputData->externFunction.arguments;
        unsigned int returnValue;
        unsigned int inputArgumentIndex = 0;
        unsigned int outputArgumentIndex = inputData->externFunction.callInterface.nargs - 1;

        buffer_t* input_buffer = inputData->parameters["input"].get_buffer().raw_buffer();
        buffer_t* output_buffer = functionData->output->buffer.raw_buffer();
        *((buffer_t**)arguments[inputArgumentIndex]) = input_buffer;
        *((buffer_t**)arguments[outputArgumentIndex]) = output_buffer;

        ffi_call(&inputData->externFunction.callInterface,
                 FFI_FN(inputData->externFunction.address),
                 &returnValue, arguments);
    } else {
        if (functionData->output->isPacked) {
            set_packed(inputData->function.output_buffer());
        }

        if (!inputData->jitCompiled) {
            // FIXME: workaround bug in Halide/LLVM where if multiple Halide::Func
            // are being JIT compiled in parallel threads a segfault will occur
            s_jitLock.lock();
            inputData->function.compile_jit();
            s_jitLock.unlock();
            inputData->jitCompiled = true;
        }
        inputData->function.realize(functionData->output->buffer);
    }
}



HalideInputData::HalideInputData(QString name)
{
    Halide::Internal::GeneratorParamValues params;
    this->generator = Halide::Internal::GeneratorRegistry::create(name.toStdString(), params);
    this->externFunction.address = dlsym(NULL, QString("halide_function_%1").arg(name).toUtf8().constData());
    this->hasExternFunction = (this->externFunction.address != NULL);

    Q_FOREACH(Halide::Internal::Parameter parameter, this->generator->get_filter_parameters()) {
        QString name = QString::fromStdString(parameter.name());
        this->parameters[name] = parameter;
    }

    if (this->hasExternFunction) {
        if (!buildExternFunction()) {
            qWarning() << QString("Unable to prepare call interface for %1").arg(name);
        }
    } else {
        Halide::Target jit_target = Halide::get_jit_target_from_environment();
        this->generator->target.set(jit_target);
        this->function = this->generator->build();
    }
}

HalideInputData::~HalideInputData()
{
    if (this->hasExternFunction) {
        destroyExternFunction();
    }
}


bool HalideInputData::buildExternFunction()
{
    const std::vector<Halide::Argument>& arguments = this->generator->get_filter_arguments();
    unsigned int nargs = arguments.size() + 1;
    this->externFunction.arguments = (void**)calloc(nargs, sizeof(void*));
    this->externFunction.argumentsTypes = (ffi_type**)calloc(nargs, sizeof(ffi_type*));

    unsigned int i = 0;
    Q_FOREACH(const Halide::Argument& argument, arguments) {
        if (argument.is_buffer) {
            // buffer_t* argument
            this->externFunction.argumentsTypes[i] = &ffi_type_pointer;
            this->externFunction.arguments[i] = (buffer_t**)calloc(1, sizeof(buffer_t*));
        } else {
            this->externFunction.argumentsTypes[i] = parameterTypeToFFIType(argument.type);
            this->externFunction.arguments[i] = NULL;
        }
        i++;
    }

    // last argument is always the output (buffer_t*)
    this->externFunction.argumentsTypes[nargs-1] = &ffi_type_pointer;
    this->externFunction.arguments[nargs-1] = (buffer_t**)calloc(1, sizeof(buffer_t*));

    ffi_status status = ffi_prep_cif(&this->externFunction.callInterface,
                                     FFI_DEFAULT_ABI, nargs, &ffi_type_sint,
                                     this->externFunction.argumentsTypes);
    return (status == FFI_OK);
}

void HalideInputData::destroyExternFunction()
{
    const std::vector<Halide::Argument>& arguments = this->generator->get_filter_arguments();

    unsigned int i = 0;
    Q_FOREACH(const Halide::Argument& argument, arguments) {
        if (argument.is_buffer) {
            free(this->externFunction.arguments[i]);
        }
        i++;
    }

    // last argument is always the output (buffer_t*)
    unsigned int nargs = externFunction.callInterface.nargs;
    free(this->externFunction.arguments[nargs-1]);

    free(this->externFunction.arguments);
    free(this->externFunction.argumentsTypes);
}

void HalideInputData::setParameter(Halide::Internal::Parameter& parameter, unsigned int index, const QVariant& value)
{
    if (parameter.is_buffer()) {
        HalideImage* halideImage = value.value<HalideImage*>();
        if (halideImage) {
            setBufferParameter(parameter, halideImage->data());
        } else {
            setBufferParameter(parameter, QSharedPointer<HalideImageData>());
        }
    } else {
        setScalarParameterValue(parameter, value);
    }

    if (this->hasExternFunction) {
        if (parameter.is_buffer()) {
            if (parameter.get_buffer().defined()) {
                *((buffer_t**)this->externFunction.arguments[index]) = parameter.get_buffer().raw_buffer();
            } else {
                *((buffer_t**)this->externFunction.arguments[index]) = NULL;
            }
        } else {
            this->externFunction.arguments[index] = parameter.get_scalar_address();
        }
    }
}

void HalideInputData::setBufferParameter(Halide::Internal::Parameter& parameter, QSharedPointer<HalideImageData> imageData)
{
    if (imageData) {
        if (imageData->isPacked) {
            set_packed(parameter);
        }
        parameter.set_buffer(imageData->buffer);
    } else {
        parameter.set_buffer(Halide::Buffer());
    }
    // keep a reference to the image backing buffer parameters
    this->imageDataCache[QString::fromStdString(parameter.name())] = imageData;
}

void HalideInputData::setBufferParameter(QString parameterName, QSharedPointer<HalideImageData> imageData)
{
    if (this->parameters.contains(parameterName)) {
        this->setBufferParameter(this->parameters[parameterName], imageData);
    }
}

QSharedPointer<HalideImageData> HalideInputData::getBufferParameter(QString parameterName)
{
    return this->imageDataCache.value(parameterName);
}

void HalideInputData::setScalarParameterValue(Halide::Internal::Parameter& parameter, QVariant value) const
{
    Halide::Type type = parameter.type();
    if (type.is_float()) {
        parameter.set_scalar<float>(value.value<float>());
    } else if (type.is_int()) {
        parameter.set_scalar<int>(value.value<int>());
    } else if (type.is_bool()) {
        parameter.set_scalar<bool>(value.value<bool>());
    } else if (type.is_uint()) {
        parameter.set_scalar<unsigned int>(value.value<unsigned int>());
    }
}

ffi_type* HalideInputData::parameterTypeToFFIType(Halide::Type type) const
{
    ffi_type* ffiType = 0;
    if (type.is_float()) {
        ffiType = &ffi_type_float;
    } else if (type.is_int()) {
        ffiType = &ffi_type_sint;
    } else if (type.is_uint()) {
        ffiType = &ffi_type_uint;
    } else if (type.is_bool()) {
        ffiType = (sizeof(bool) == 1 ? &ffi_type_uint8 : &ffi_type_uint32);
    }
    return ffiType;
}
