#ifndef HALIDETRANSFORMWORKER_H
#define HALIDETRANSFORMWORKER_H

#include <QtCore/QObject>
#include <QtCore/QHash>
#include <QtCore/QSharedPointer>
#include <QtCore/QThread>
#include <QtCore/QMutex>
#include <Halide.h>
#include <ffi.h>

#include "halide_image.h"

struct HalideInputData {
    std::unique_ptr<Halide::Internal::GeneratorBase> generator;
    Halide::Func function;
    bool jitCompiled = false;
    QHash<QString, QSharedPointer<HalideImageData> > imageDataCache;
    QHash<QString, Halide::Internal::Parameter> parameters;
    bool hasExternFunction = false;
    struct ExternFunction {
        ffi_cif callInterface;
        void* address = 0;
        void** arguments = 0;
        ffi_type** argumentsTypes = 0;
    } externFunction;

    HalideInputData(QString name);
    ~HalideInputData();
    bool buildExternFunction();
    void destroyExternFunction();
    void setParameter(Halide::Internal::Parameter& parameter, unsigned int index, const QVariant& value);
    void setBufferParameter(Halide::Internal::Parameter& parameter, QSharedPointer<HalideImageData> imageData);
    void setBufferParameter(QString parameterName, QSharedPointer<HalideImageData> imageData);
    QSharedPointer<HalideImageData> getBufferParameter(QString parameterName);
    void setScalarParameterValue(Halide::Internal::Parameter& parameter, QVariant value) const;
    ffi_type* parameterTypeToFFIType(Halide::Type type) const;
};

struct HalideFunctionData {
    QSharedPointer<HalideInputData> input;
    QSharedPointer<HalideImageData> output;
    bool outputContentsValid = false;
};

Q_DECLARE_METATYPE(QSharedPointer<HalideFunctionData>)

class HalideFunction;
class HalideTransform;
class HalideTransformWorker : public QObject
{
    Q_OBJECT

public:
    explicit HalideTransformWorker(HalideTransform* worker);
    ~HalideTransformWorker();

    void compute(HalideImage* inputImage, QList<HalideFunction*> functions, bool qimageRequired);
    virtual bool isReady() { return true; }
    virtual QSharedPointer<HalideImageData> prepareOutputData(QSharedPointer<HalideImageData> oldOutputData, QSize outputSize) = 0;
    virtual void preRealizeChainFunctions() {}
    virtual void postRealizeChainFunctions() {}
    void pruneCache(QList<HalideFunction*> validFunctions);

Q_SIGNALS:
    void updateRequired();
    void computeFinished(QSharedPointer<HalideImageData> result);

protected:
    QList<HalideFunction*> functionsToRecompute(QList<HalideFunction*> functions, QSharedPointer<HalideImageData>& lastOutput) const;
    QList< QSharedPointer<HalideFunctionData> > prepareFunctionsData(QList<HalideFunction*> functions);
    Q_SLOT void realizeChainFunctions(QList< QSharedPointer<HalideFunctionData> > functionsData, bool qimageRequired);
    void realizeFunction(QSharedPointer<HalideFunctionData> functionData);

    HalideTransform* m_transform;
    QHash<HalideFunction*, QSharedPointer<HalideFunctionData> > m_cache;
    QThread* m_thread;
    QSharedPointer<HalideImageData> m_outputImage;
    static QMutex s_jitLock;
};

#endif // HALIDETRANSFORMWORKER_H

