# Halide supports the following:
#
# Arches:
# "x86", "arm", "pnacl", "mips"
#
# OSes:
# "linux", "windows", "osx", "android", "ios", "nacl"
#
# Features:
# "jit", "debug", "no_asserts", "no_bounds_query",
# "sse41", "avx", "avx2", "fma", "fma4", "f16c",
# "armv7s", "no_neon",
# "cuda", "cuda_capability_30", "cuda_capability_32", "cuda_capability_35", "cuda_capability_50",
# "opencl", "cl_doubles",
# "opengl",
# "egl",
# "glx",
#
# Halide target format looks like: arch-bits-os-feature1-feature2-...
# Examples:
# x86-64-linux-sse41-avx-fma-f16c
# arm-32-linux

# computing TARGET_OS, TARGET_PROCESSOR and TARGET
set(TARGET_OS linux)

if(CMAKE_CROSSCOMPILING AND ${CMAKE_SYSTEM_PROCESSOR} STREQUAL ${CMAKE_HOST_SYSTEM_PROCESSOR})
    set(TARGET_PROCESSOR $ENV{DEB_HOST_ARCH_CPU}-$ENV{DEB_HOST_ARCH_BITS})
else()
    set(TARGET_PROCESSOR ${CMAKE_SYSTEM_PROCESSOR})
endif()

string(REPLACE "_" "-" TARGET_PROCESSOR ${TARGET_PROCESSOR})
string(TOLOWER ${CMAKE_SYSTEM_NAME} TARGET_OS)
set(TARGET ${TARGET_PROCESSOR}-${TARGET_OS})

# When doing a native build, create linkable generator object files (.o)
# for all the target architectures (TARGET_ARCHES); they are created in
# the source tree because they will be reused by subsequent non-native
# builds (when cross compiling)
set(TARGET_ARCHES
    x86-64
    arm-32
#    pnacl
#    mips
)

function(compile_generators TARGET_DIR TARGET_PROJECT_NAME)
    file(GLOB GENERATORS RELATIVE "${TARGET_DIR}" "${TARGET_DIR}/generator_*.cpp")

    if(NOT CMAKE_CROSSCOMPILING)
        foreach(GENERATOR_SRC ${GENERATORS})
            string(REPLACE ".cpp" "" GENERATOR_NAME ${GENERATOR_SRC})
            string(REPLACE "generator_" "halide_function_" FUNCTION_NAME ${GENERATOR_NAME})
            message("Will produce generator binaries for: " ${GENERATOR_NAME})

            # create executable for the generator
            set(GENERATOR_EXEC generator_${GENERATOR_NAME})
            add_executable(${GENERATOR_EXEC} ${TARGET_DIR}/GenGen.cpp
                                             ${TARGET_DIR}/${GENERATOR_SRC})
            qt5_use_modules(${GENERATOR_EXEC} Gui)
            target_link_libraries(${GENERATOR_EXEC} ${HALIDE_LIBRARIES})

            foreach(CROSS_TARGET_ARCH ${TARGET_ARCHES})
                set(CROSS_TARGET_OS ${TARGET_OS})
                set(CROSS_TARGET ${CROSS_TARGET_ARCH}-${CROSS_TARGET_OS})
                set(GENERATOR_OBJECT_NAME ${GENERATOR_NAME}_${CROSS_TARGET})
                set(GENERATOR_OBJECT_FILE ${GENERATOR_OBJECT_NAME}.o)
                if(${CROSS_TARGET} STREQUAL ${TARGET})
                    set(HALIDE_TARGET "host")
                else()
                    set(HALIDE_TARGET ${CROSS_TARGET})
                endif()

                # invoke the executable created prior
                add_custom_command(OUTPUT ${CMAKE_CURRENT_SOURCE_DIR}/${GENERATOR_OBJECT_FILE}
                    DEPENDS ${GENERATOR_EXEC}
                    COMMAND ${CMAKE_CURRENT_BINARY_DIR}/${GENERATOR_EXEC} -f ${FUNCTION_NAME} -b ${GENERATOR_OBJECT_NAME} -o . target=${HALIDE_TARGET}
                    WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR}
                    COMMENT "Producing generator object file for target: " ${HALIDE_TARGET} " into " ${GENERATOR_OBJECT_FILE}
                    VERBATIM
                )
                add_custom_target(${GENERATOR_OBJECT_FILE}_target
                                  DEPENDS ${CMAKE_CURRENT_SOURCE_DIR}/${GENERATOR_OBJECT_FILE}
                )
                add_dependencies(${TARGET_PROJECT_NAME} ${GENERATOR_OBJECT_FILE}_target)
            endforeach()
        endforeach()
    endif(NOT CMAKE_CROSSCOMPILING)

    message("Will link generators for target architecture: " ${TARGET})
    foreach(GENERATOR_SRC ${GENERATORS})
        string(REPLACE ".cpp" "" GENERATOR_NAME ${GENERATOR_SRC})
        set(GENERATOR_OBJECT_NAME ${GENERATOR_NAME}_${TARGET})
        set(GENERATOR_OBJECT_FILE ${GENERATOR_OBJECT_NAME}.o)
        message(" - " ${GENERATOR_NAME})

        # add object generated by invoking the executable to a list
        set(GENERATORS_OBJECTS ${GENERATORS_OBJECTS} ${CMAKE_CURRENT_SOURCE_DIR}/${GENERATOR_OBJECT_FILE})
    endforeach()

    set(${TARGET_PROJECT_NAME}_GENERATORS_OBJECTS ${GENERATORS_OBJECTS} PARENT_SCOPE)
endfunction()
