import QtQuick 2.0
import QtTest 1.0
import QtHalide 0.1
import QtHalideTestHelper 0.1

Item {
    width: 200
    height: 200

    // coverage: exercise access to HalideFunctionListModel::roleNames
    ListView {
        anchors.fill: parent
        id: list
        delegate: Text {
            text: modelData.name
            height: 60
            width: parent.width
        }
    }

    HalideTransform {
        id: transform
        live: true
        input: input
    }

    HalideImage {
        id: input
        source: "in.png"
    }

    HalideImageRenderer {
        id: output
        anchors.centerIn: parent
        width: input.size.width
        height: input.size.height
        image: transform.output
        clip: true
    }

    HalideFunction {
        id: red
        objectName: "red"
        name: "channel"
        property HalideImage input
        property int channel: 0
    }

    HalideFunction {
        id: green
        objectName: "green"
        name: "channel"
        property HalideImage input
        property int channel: 1
    }

    HalideFunction {
        id: blue
        objectName: "blue"
        name: "channel"
        property HalideImage input
        property int channel: 2
    }

    HalideTransformSignalSpy {
        id: spy
        halideTransform: transform
        test: test
    }

    ComparePixelsTestCase {
        id: test
        name: "FunctionsListModelTest"
        when: windowShown

        compareTarget: output

        function init() {
            spy.ensureNoPendingUpdate();
            spy.clear();
        }

        function cleanup() {
            transform.functions = [];
            red.enabled = true;
            green.enabled = true;
            blue.enabled = true;
            spy.clear();
        }

        function test_append() {
            compare(transform.functions.count, 0, "list not empty at start");
            transform.functions = [red, green, blue];
            compare(transform.functions.count, 3, "list count not updated correctly");
            compare(transform.functions.get(0).name, "channel", "function name is not correct");

            compare(transform.functions.get(0).objectName, "red", "function object name is not correct");
            compare(transform.functions.get(1).objectName, "green", "function object name is not correct");
            compare(transform.functions.get(2).objectName, "blue", "function object name is not correct");
        }

        function test_clear() {
            transform.functions = [red, green, blue];
            transform.functions = [];
            compare(transform.functions.count, 0, "list not empty after clear");
        }

        function test_enable_after_assign() {
            transform.functions = [blue, red, green];
            spy.waitForUpdate();

            compareAllPixelsWithColor(255, 255, 255, 255, "output does not match expected output");

            red.enabled = false;
            spy.waitForUpdate();
            compareAllPixelsWithColor(128, 255, 255, 255, "output does not match expected when first function disabled");

            red.enabled = true;
            spy.waitForUpdate();
            compareAllPixelsWithColor(255, 255, 255, 255, "output does not match expected output");

            green.enabled = false;
            spy.waitForUpdate();
            blue.enabled = false;
            spy.waitForUpdate();
            compareAllPixelsWithColor(255, 128, 128, 255, "leaving only the first filter produce incorrect output");
        }

        function test_add_disabled_then_enable() {
            red.enabled = false;
            transform.functions = [red];
            spy.waitForUpdate();
            red.enabled = true;
            spy.waitForUpdate();
            compareAllPixelsWithColor(255, 128, 128, 255, "output does not match expected output");
        }

        function test_all_disabled() {
            red.enabled = false;
            green.enabled = false;
            transform.functions = [red, green];
            compareAllPixelsWithColor(128, 128, 128, 255, "output is not equal to input when no function is enabled");
        }

        function test_append_null() {
            ignoreWarning("HalideFunctionListModel::insert: tried to insert a null function")
            transform.functions.append(null);
            compare(transform.functions.count, 0, "appending null should not add items to the list");
        }

        function test_assign_list_with_null() {
            ignoreWarning("HalideFunctionListModel::insert: tried to insert a null function");
            transform.functions = [red, null, blue];
            compare(transform.functions.count, 2, "wrong number of items assigned when list contains null");
            compare(transform.functions.get(0).objectName, "red", "function object name is not correct");
            compare(transform.functions.get(1).objectName, "blue", "function object name is not correct");
        }

        function test_insert() {
            transform.functions = [green, blue];
            transform.functions.insert(1, red);
            compare(transform.functions.count, 3, "function was not added when inserted");
            compare(transform.functions.get(1).objectName, "red", "function was not added in the right place when inserted below bounds");
        }

        function test_insert_null() {
            ignoreWarning("HalideFunctionListModel::insert: tried to insert a null function");
            transform.functions = [green, blue];
            transform.functions.insert(1, null);
            compare(transform.functions.count, 2, "invalid value should not have been inserted");
        }

        function test_insert_below_bounds() {
            ignoreWarning("HalideFunctionListModel::insert: index passed is out of bounds.")
            transform.functions = [green, blue];
            transform.functions.insert(-1, red);
            compare(transform.functions.count, 3, "function was not added when inserted out of bounds");
            compare(transform.functions.get(0).objectName, "red", "function was not added at start when inserted below bounds");
        }

        function test_insert_above_bounds() {
            ignoreWarning("HalideFunctionListModel::insert: index passed is out of bounds.")
            transform.functions = [green, blue];
            transform.functions.insert(5, red);
            compare(transform.functions.count, 3, "function was not added when inserted out of bounds");
            compare(transform.functions.get(2).objectName, "red", "function was not added at end when inserted below bounds");
        }

        function test_remove_index() {
            transform.functions = [red];
            transform.functions.remove(0);
            compare(transform.functions.count, 0, "function was not removed");
        }

        function test_remove_function() {
            transform.functions = [red, green, blue];
            transform.functions.remove(green);
            compare(transform.functions.count, 2, "function was not removed");
            compare(transform.functions.get(0).objectName, "red", "wrong function was removed");
            compare(transform.functions.get(1).objectName, "blue", "wrong function was removed");
        }

        function test_remove_function_null() {
            transform.functions = [red, green, blue];
            transform.functions.remove(null);
            compare(transform.functions.count, 3, "list was altered when asked to remove null function");
        }

        function test_move() {
            transform.functions = [red, green, blue];
            transform.functions.move(2, 0);
            compare(transform.functions.count, 3, "function count was not altered while moving");
            compare(transform.functions.get(0).objectName, "blue", "wrong function was moved");
            compare(transform.functions.get(1).objectName, "red", "wrong function was moved");
            compare(transform.functions.get(2).objectName, "green", "wrong function was moved");
        }

        function test_move_out_of_bounds() {
            transform.functions = [red, green, blue];
            transform.functions.move(-1, 20);
            compare(transform.functions.get(0).objectName, "red", "wrong function was moved");
            compare(transform.functions.get(1).objectName, "green", "wrong function was moved");
            compare(transform.functions.get(2).objectName, "blue", "wrong function was moved");
        }

        function test_access_out_of_bounds() {
            transform.functions = [red, green, blue];
            compare(transform.functions.get(-1), null, "wrong value returned when accessing out of bounds");
            compare(transform.functions.get(100), null, "wrong value returned when accessing out of bounds");
        }

        function test_access_role_name() {
            transform.functions = [red, green, blue];
            list.model = transform.functions;
        }
    }
}

