import QtQuick 2.0
import QtTest 1.0
import QtHalide 0.1
import QtHalideTestHelper 0.1

Item {
    id: main
    width: 200
    height: 200

    Rectangle {
        anchors.fill: parent
        color: "black"
    }

    HalideImageRenderer {
        id: output
        anchors.centerIn: parent
        width: input.size.width
        height: input.size.height
        image: input
        clip: true
    }

    HalideImage {
        id: input

        function setSourceAndCheckStatus(source, expected) {
            input.source = source;
            test.tryCompare(input, "status", expected, 5000, "input did not reach the expected status");
        }
    }

    HalideImage {
        id: yellow
        source: "yellow.png"
    }

    TestHelper { id: helper }

    ComparePixelsTestCase {
        id: test
        name: "ImageRendererTest"
        when: windowShown

        compareTarget: output

        function cleanup() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            input.sourceSize.width = 0
            input.sourceSize.height = 0
            yellow.source = "yellow.png"
            tryCompare(yellow, "status", Image.Ready, 5000, "failed to reset source of yellow test image")
            output.image = input
            output.width = input.size.width
            output.height = input.size.height
        }

        function test_assign_render() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            compareAllPixelsWithColor(128, 128, 128, 255, "image not rendered");
        }

        function test_reassign_render() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            input.setSourceAndCheckStatus("yellow.png", Image.Ready);
            compareAllPixelsWithColor(255, 255, 0, 255, "image not rendered");
        }

        function test_null_render() {
            output.image = null;
            compare(output.width, 0);
            compare(output.height, 0);
        }

        function test_setData_render() {
            skip("QtHalideTestHelper does not have setImageData for now")
            helper.setImageData(input, yellow);
            compareAllPixelsWithColor(255, 255, 0, 255, "image not rendered");
        }

        function test_left_h_align() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.horizontalAlignment = HalideImageRenderer.AlignLeft;
            output.width = input.size.width + 2;
            comparePixelWithColor(0, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(9, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(10, 0, 0, 0, 0, 255, "pixel outside image is not empty");
        }

        function test_right_h_align() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.horizontalAlignment = HalideImageRenderer.AlignRight;
            output.width = input.size.width + 2;
            comparePixelWithColor(1, 0, 0, 0, 0, 255, "pixel outside image is not empty");
            comparePixelWithColor(2, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(11, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
        }

        function test_center_h_align() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.horizontalAlignment = HalideImageRenderer.AlignHCenter;
            output.width = input.size.width + 2;
            comparePixelWithColor(0, 0, 0, 0, 0, 255, "pixel outside image is not empty");
            comparePixelWithColor(1, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(10, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(11, 0, 0, 0, 0, 255, "pixel outside image is not empty");
        }

        function test_top_v_align() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.verticalAlignment = HalideImageRenderer.AlignTop;
            output.height = input.size.height + 2;
            comparePixelWithColor(0, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(0, 9, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(0, 10, 0, 0, 0, 255, "pixel outside image is not empty");
        }

        function test_bottom_v_align() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.verticalAlignment = HalideImageRenderer.AlignBottom;
            output.height = input.size.height + 2;
            comparePixelWithColor(0, 1, 0, 0, 0, 255, "pixel outside image is not empty");
            comparePixelWithColor(0, 2, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(0, 11, 128, 128, 128, 255, "pixel inside image has unexpected value");
        }

        function test_center_v_align() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.verticalAlignment = HalideImageRenderer.AlignVCenter;
            output.height = input.size.height + 2;
            comparePixelWithColor(0, 0, 0, 0, 0, 255, "pixel outside image is not empty");
            comparePixelWithColor(0, 1, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(0, 10, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(0, 11, 0, 0, 0, 255, "pixel outside image is not empty");
        }

        function test_fill_mode_fit() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.verticalAlignment = HalideImageRenderer.AlignVCenter;
            output.horizontalAlignment = HalideImageRenderer.AlignHCenter;
            output.fillMode = HalideImageRenderer.PreserveAspectFit;
            output.height = input.size.height - 2;

            compare(output.fillMode, HalideImageRenderer.PreserveAspectFit);
            compare(output.verticalAlignment, HalideImageRenderer.AlignVCenter);
            compare(output.horizontalAlignment, HalideImageRenderer.AlignHCenter);

            comparePixelWithColor(0, 0, 0, 0, 0, 255, "pixel outside image is not empty");
            comparePixelWithColor(1, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(8, 0, 128, 128, 128, 255, "pixel inside image has unexpected value");
            comparePixelWithColor(9, 0, 0, 0, 0, 255, "pixel outside image is not empty");
        }

        function test_fill_mode_crop() {
            input.setSourceAndCheckStatus("in.png", Image.Ready);
            output.verticalAlignment = HalideImageRenderer.AlignVCenter;
            output.horizontalAlignment = HalideImageRenderer.AlignHCenter;
            output.fillMode = HalideImageRenderer.PreserveAspectCrop;
            output.height = input.size.height - 2;

            compareAllPixelsWithColor(128, 128, 128, 255, "output should be entirely filled");
        }
    }
}

