import QtQuick 2.0
import QtTest 1.0
import QtHalide 0.1
import QtHalideTestHelper 0.1

Item {
    id: main
    width: 200
    height: 200

    HalideTransform {
        id: transform
        live: true
    }

    HalideImageRenderer {
        id: output
        anchors.centerIn: parent
        width: input.size.width
        height: input.size.height
        image: transform.output
        clip: true
    }

    HalideFunction {
        id: red
        name: "channel"
        property HalideImage input
        property int channel: 0
    }

    HalideImage {
        id: input
        source: "in.png"
    }

    HalideImage {
        id: input2
        source: "yellow.png"
    }

    HalideTransformSignalSpy {
        id: spy
        halideTransform: transform
        test: test
    }

    ComparePixelsTestCase {
        id: test
        name: "TransformTest"
        when: windowShown

        compareTarget: output

        function cleanup() {
            transform.functions = [];
            transform.input = null;
            transform.live = true;
            red.channel = 0;
            spy.clear();
        }

        // when removing all functions ensure output is same as input
        function test_output_reset_when_no_functions() {
            transform.input = input;
            transform.functions = [red];
            spy.waitForUpdate();
            transform.functions = [];
            compareAllPixelsWithColor(128, 128, 128, 255, "unexpected output present when removing all functions");

            compare(transform.updating, false, "Updating still true even when no updates should be happening"); // coverage
        }

        // verify rendering correct when adding function with input already set
        function test_functions_applied_with_input() {
            transform.input = input;
            compare(spy.count, 0, "Should have received no updated while transform has no functions");
            transform.functions = [red];
            spy.waitForUpdate();
            compareAllPixelsWithColor(255, 128, 128, 255, "filters not applied");
        }

        // verify no output when there is no function
        function test_functions_applied_without_input() {
            transform.functions = [red];
            compare(spy.count, 0, "Should have received no updates while transform has no input");
            transform.input = input;
            spy.waitForUpdate();
            compareAllPixelsWithColor(255, 128, 128, 255, "filters not applied");
        }

        // verify that turning live on and off turns property changed updates on and off
        function test_live_signals() {
            transform.input = input;
            transform.functions = [red];
            spy.waitForUpdate();
            red.channel = 2;
            spy.waitForUpdate();

            transform.live = false;
            compare(transform.live, false, "live property not set correctly"); //coverage

            spy.clear();
            red.channel = 1;
            compare(spy.count, 0, "Updates should not be fired while not live");

            transform.live = true;
            spy.waitForUpdate();
            red.channel = 2;
            spy.waitForUpdate();
        }

        // verify that turning live on and off works as expected
        function test_live() {
            transform.live = false;
            transform.input = input;
            transform.functions = [red];
            compareAllPixelsWithColor(128, 128, 128, 255, "unexpected output present before going live");

            transform.live = true;
            spy.waitForUpdate();
            compareAllPixelsWithColor(255, 128, 128, 255, "filters were not applied when live");

            transform.live = false;
            compareAllPixelsWithColor(255, 128, 128, 255, "output changed after going off live");
        }

        // verify changing input will correctly be done
        function test_switching_input() {
            transform.functions = [];
            transform.input = input;
            transform.input = input2;
            compareAllPixelsWithColor(255, 255, 0, 255, "switching input is not appearing on output");
        }

        // verify that setting the functions to an invalid type will not crash and do nothing
        function test_set_functions_invalid() {
            ignoreWarning("HalideTransform::setFunctions received an argument of the wrong type. Expected a QVariant(QVariantList, (QVariant(HalideFunction*), ... ) )");
            transform.functions = "hello";
            compare(transform.functions.count, 0, "function list is not empty after invalid set");
        }
    }
}

